# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Module containing Design interface components."""
from copy import deepcopy
from typing import TYPE_CHECKING, List, Union
#from typing import Any, Optional, TypeVar, Dict as Dict_, Iterable
from qiskit_metal.qlibrary.core._parsed_dynamic_attrs import is_ipython_magic
from qiskit_metal import logger
from qiskit_metal import config

if not config.is_building_docs():
    # Only import QComponent if the docs are NOT being built
    from ..qlibrary.core import QComponent

if TYPE_CHECKING:
    # For linting typechecking, import modules that can't be loaded here under normal conditions.
    # For example, I can't import QDesign, because it requires QComponent first. We have the
    # chicken and egg issue.
    from .design_base import QDesign


class Components:
    """This is a user interface for the design._components dict.

    The keys are unique integers, however, this interface allows user to
    treat the keys as strings.
    """

    def __init__(self, design: 'QDesign'):
        """Set up variables and logger which are used to emulate a dict which
        is referencing design._components.

        Args:
            design (QDesign): Need to have a Qdesign class so this class can
                              reference design._components.
        """
        self._design = design
        self.logger = logger  # type: logging.Logger
        self.components = design._components
        self.name_list = list()
        self.name_list_idx = 0

    def __len__(self) -> int:
        """Give number of components in design.

        Returns:
            int: Total number of components registered within the design.
        """
        return len(self.components)

    def get_list_ints(self, component_names: List[str]) -> List[int]:
        """Provide corresponding ints to be used as keys for dict:
        design._components, when list of names is provided.

        Args:
            component_names (List[str]): Names of components which user wants to know
                                         the int to be used as a key.

        Returns:
            List[int]: Corresponding ints that user can use as keys into design._components
        """
        component_ints = [self.find_id(item) for item in component_names]
        return component_ints

    def find_id(self, name: str, quiet: bool = False) -> int:
        """Find id of component.  The id is the key for a dict which holds all
        of the components within design.

        Args:
            name (str): Text name of component.  The name is assumed to be unique.
            quiet (bool): Allow warning messages to be generated.

        Returns:
            int: Key to use in  _components.
            If 0 is returned it means the name is not in dict.

        Raises:
            AttributeError: The given name is a magic method not in the dictionary
        """

        if name in self._design.name_to_id:
            component_id = self._design.name_to_id[name]
            return component_id
        if not is_ipython_magic(name):
            # Name not registered, not in cache for components' names.
            # IPython checking methods
            # https://github.com/jupyter/notebook/issues/2014
            if not quiet:
                self.logger.warning(
                    f'In Components.find_id(), the name={name} is not used in design._components'
                )
            return 0

        raise AttributeError(name)

    # def is_name_used(self, new_name: str) -> int:
    #     """Check to see if name being used in components.

    #      Args:
    #          new_name (str): name to check
    #
    #      Returns:
    #          int: If the name does not exist, 0 is returned, otherwise the
    #          component-id of component which is already using the name.
    #     """

    #     all_names = [(value.name, key)
    #                  for (key, value) in self.components.items()]
    #     search_result = [
    #         item for item in all_names if new_name == item[0]]
    #     if len(search_result) != 0:
    #         self.logger.warning(
    #             f'Called interface_components, '
    #             f' component_id({search_result[0][0]}, id={search_result[0][1]})'
    #             f' is already using name={new_name}.')
    #         return search_result[0][1]
    #     else:
    #         return 0

    def __getitem__(self,
                    name: str,
                    quiet: bool = False) -> Union[None, 'QComponent']:
        """Get the QComponent based on string name vs the unique id of
        QComponent.

        Args:
            name (str): Name of component.
            quiet (bool): Allow warning messages to be generated.

        Returns:
            QComponent: Class which describes the component.  None if
            name not found in design._components.

        Raises:
            AttributeError: The given name is a magic method not in the dictionary
        """
        component_id = int(self.find_id(name))
        if component_id:
            return self._design._components[component_id]
        else:
            # IPython checking methods
            # https://github.com/jupyter/notebook/issues/2014
            if not is_ipython_magic(name):
                if not quiet:
                    self.logger.warning(
                        f'In Components.__getitem__, name={name} is not '
                        f'registered in the design class. Return '
                        f'None for QComponent.')
                return None
            else:
                raise AttributeError(name)

            return None

        raise AttributeError(name)

    def __setitem__(self, name: str, value: 'QComponent'):
        """Replace QComponent for an existing name. Use this at your own risk.
        There are netids used for pins within a component.  The netids are used
        in the net_info table and qgeometry tables.

        Args:
            name (str): Name of QComponent.  If not in design._components,
                        then will be added to dict.
                        If in dict, the value(QComponent) will replace existing QComponent.
            value (QComponent): QComponent to add under the given name
        """
        if not isinstance(value, QComponent):
            self.logger.warning(
                f'The value is NOT a QComponent.  Nothing has been assigned to name={name}'
            )
            return

        component_id = self.find_id(name)
        if component_id:
            self.logger.debug(
                f'The name={name} already exists in design._components.  '
                f'A component_id={component_id} will be replaced.')
            self.components[component_id] = deepcopy(value)
        else:
            self.logger.warning(
                f'Usualy new components are added to design during init.  '
                f'The name={name} is not in design._components, and added as a new component.'
            )
            value.name = name
            value._add_to_design()

    def __getattr__(self, name: str) -> Union['QComponent', None]:
        """Provide same behavior as __getitem__.

        Args:
            name (str): Name of component used to find the QComponent in
                    design._components dict, vs using unique int id.

        Returns:
            QComponent: Class which describes the component. None if
                        name not found in design._components.
        """
        quiet = True
        return self.__getitem__(name, quiet)

    # def __setattr__(self, name: str, value: 'QComponent'):
    #     """Provide same behavior as __setitem__.

    #     Args:
    #         name (str): Name of component used to find the QComponent in
    #                   design._components dict, vs using unique int id.
    #         value (QComponent): Component with the name used in arguments.
    #     """
    #     pass
    #     #self.__setitem__(name, value)

    def __contains__(self, item: str) -> int:
        """Look for item in design._components in the value.

        Args:
            item (str): Name in the value of design._components.

        Returns:
            int: 0 if item is not in design._components.name otherwise an int which can be used as
                 key for design._components[]
        """
        if not isinstance(item, str):
            # self.logger.debug(f'Search with string in __contains__ {item}.')
            return 0
        quiet = True
        return self.find_id(item, quiet)

    # def __delitem__(self, name: str):
    #     """Will delete component from design class along with deleting the
    #      net_info and element tables.

    #     Args:
    #         name (str): Name of component to delete from design._components.
    #     """
    #     component_id = self.is_name_used(name)
    #     self._design.delete_component(component_id)

    def __repr__(self) -> str:
        """Print the design._component dict.

        Returns:
            str : String to print design._component dict.
        """

        return str(self._design._components.__repr__())

    #     #     def __repr__(self):
    #     #         # make sure to define representation for print purpose
    #     #         # Why every class needs one?  https://dbader.org/blog/python-repr-vs-str
    #     #         return str(self.__actual_place_I_store_components__.give_me_repr())

    def __dir__(self) -> List:
        """Provide all the names in design._components.

        Returns:
            list: List of all the names used in design._components.
        """
        all_names = [(value.name) for (key, value) in self.components.items()]
        return all_names

        #     def __dir__(self):
        #         # For autocompletion
        #         return list(self.__get_dict__().keys())

    def __iter__(self) -> iter:
        """Give iterator for design._components.

        Returns:
            iter: for design._components , the keys are names of the components.
        """
        for value in self._design._components.values():
            new_key = value.name
            yield new_key

    def items(self) -> list:
        """Get a list of all the items.

        Returns:
            list: List of (key, value) pairs.
        """
        all_items = [
            (value.name, value) for (key, value) in self.components.items()
        ]
        return all_items

    def values(self) -> list:
        """Get the list of all the values.

        Returns:
            list: List of just the values.
        """
        all_items = [value for (key, value) in self.components.items()]

        return all_items

    def keys(self) -> list:
        """Get the list of all the keys.

        Returns:
            list: List of just the keys.
        """
        all_items = [value.name for (key, value) in self.components.items()]

        return all_items

        #     #### Down the line for serialization and pickling. Skip for now
        #     def __getstate__(self):
        #         """
        #         Serialize the object.
        #         """
        #         # something like
        #         return self.__actual_place_I_store_components__.__getstate__()
        #     def __setstate__(self, state):
        #         """
        #         Deserialize the object.
        #         """
        #         pass



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=invalid-name
# pylint: disable=unused-import
"""File contains some config definitions.

Mostly internal.
"""
import os
import sys

from .toolbox_python.attr_dict import Dict
from ._defaults import DefaultMetalOptions, DefaultOptionsRenderer

renderers_to_load = Dict(
    hfss=Dict(path_name='qiskit_metal.renderers.renderer_ansys.hfss_renderer',
              class_name='QHFSSRenderer'),
    q3d=Dict(path_name='qiskit_metal.renderers.renderer_ansys.q3d_renderer',
             class_name='QQ3DRenderer'),
    gds=Dict(path_name='qiskit_metal.renderers.renderer_gds.gds_renderer',
             class_name='QGDSRenderer'),
    gmsh=Dict(path_name='qiskit_metal.renderers.renderer_gmsh.gmsh_renderer',
              class_name='QGmshRenderer'),
    elmer=Dict(path_name='qiskit_metal.renderers.renderer_elmer.elmer_renderer',
               class_name='QElmerRenderer'),
    aedt_q3d=Dict(
        path_name=
        'qiskit_metal.renderers.renderer_ansys_pyaedt.q3d_renderer_aedt',
        class_name='QQ3DPyaedt'),
    aedt_hfss=Dict(
        path_name=
        'qiskit_metal.renderers.renderer_ansys_pyaedt.hfss_renderer_aedt',
        class_name='QHFSSPyaedt'))
"""
Define the renderes to load. Just provide the module names here.
"""

GUI_CONFIG = Dict(
    load_metal_modules=Dict(Qubits='qiskit_metal.qlibrary.qubits',
                            TLines='qiskit_metal.qlibrary.tlines',
                            Terminations='qiskit_metal.qlibrary.terminations'),
    exclude_metal_classes=['Metal_Qubit'],
    tips=[
        'Right clicking the tree elements allows you to do neat things.',
        'You can show all connector names on the plot by clicking the connector '
        'icon in the plot toolbar.',
        'The gui and the Python code work synchronously. If you modify something '
        'in the gui, it will be reflected in your Python interpreter and vice versa. '
        'Note that the gui does not automatically refresh on all events if you update '
        'variables from the Python interpreter.',
        'Changed some object parameters? Click the <b>Rebuild</b> button in the main '
        'toolbar to recreate the polygons.',
        """<b>Log widget:</b> Right click the logger window to be able to change the log level and
        the loggers that are shown / hidden.""",
        """<b>All component widget:</b> Double click a component to zoom into it!""",
    ],
    logger=Dict(
        style=
        ".DEBUG {color: green;}\n.WARNING,.ERROR,.CRITICAL {color: red;}\n.'\
                'ERROR,.CRITICAL {font-weight: bold;}\n",
        num_lines=500,
        level='DEBUG',
        stream_to_std=False,  # stream to jupyter notebook
    ),
    main_window=Dict(
        title='Qiskit Metal — The Quantum Builder',
        auto_size=False,  # Autosize on creation of window
    ))
"""
GUI_CONFIG

**load_metal_modules**

---------------------------
Name of class folders that contain modules that will be available to be
created in the GUI

Conventions:
Assumes that the module file name is the same as the class name contained in it.
For example, provided `qiskit_metal.qubits` has `Metal_Transmon_Pocket.py`, the
gui will do
    `from qiskit_metal.qubits.Metal_Transmon_Pocket import Metal_Transmon_Pocket`


**tips**

---------------------------
Tips that the user can define to show in the gui. These rotate each time the gui is started.


**logger**

---------------------------
Logger settings


**main_window**

---------------------------
Main window defaults
"""

log = Dict(format='%(asctime)s %(levelname)s [%(funcName)s]: %(message)s',
           datefmt='%I:%M%p %Ss')
"""
A dictionary containing the log format for standard text and date/time
"""


def is_using_ipython() -> bool:
    """Check if we're in IPython.

    Returns:
        bool -- True if ran in IPython
    """
    return 'JPY_PARENT_PID' in os.environ


def is_building_docs() -> bool:
    """Return True if we are in a documentation (Sphinx) build context."""
    if os.getenv("QISKIT_METAL_DOCS_BUILD") == "1":
        return True
    # Fallback heuristic if someone builds docs without our env flag
    return "sphinx" in sys.modules

    # OLD:
    # """Checks for the existance of the .buildingdocs file which is only present
    # when building the docs.
    # Returns:
    #     bool: True if .buildingdocs exists
    # """
    # from pathlib import Path  # pylint: disable=import-outside-toplevel
    # build_docs_file = Path(__file__).parent.parent / "docs" / ".buildingdocs"
    # return Path.exists(build_docs_file)


_ipython = is_using_ipython()

####################################################################################
# USER CONFIG



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Handles the capacitive simulation of a transmon pocket model.

Convert capacitance matrices extracted from Q3D into Hamiltonian parameters
using the Duffing model. Typical input is the capacitance matrix calculated from Q3D.

Each function prints out the parameters and outputs a dictionary.
"""
# pylint: disable=invalid-name

import io
import re
from pathlib import Path
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.optimize as opt
from pint import UnitRegistry

from pyEPR.calcs.convert import Convert
from .constants import (e, h, hbar, phinot, phi0)

__all__ = [
    'Ic_from_Lj', 'Ic_from_Ej', 'Cs_from_Ec', 'transmon_props', 'chi',
    'extract_transmon_coupled_Noscillator', 'levels_vs_ng_real_units',
    'get_C_and_Ic', 'cos_to_mega_and_delta', 'chargeline_T1',
    'readin_q3d_matrix', 'readin_q3d_matrix_m', 'load_q3d_capacitance_matrix',
    'df_cmat_style_print', 'move_index_to', 'df_reorder_matrix_basis',
    'lumped_oscillator_from_path'
]


def Ic_from_Lj(Lj: float) -> float:
    """Critical current In SI units.

    Args:
        Lj (float): In Henries

    Returns:
        float: In Amps
    """
    return phi0 / Lj


def Ic_from_Ej(Ej: float) -> float:
    """Critical current In SI units.

    Args:
        Ej (float): In Joules

    Returns:
        float: In Amps
    """
    return Ej / phi0


def Cs_from_Ec(Ec: float) -> float:
    """Get total shunt capacitance from charging energy In SI units.

    Args:
        Ec (float): In Joules

    Returns:
        float: In farads
    """
    return e**2 / (2 * Ec)


def transmon_props(Ic: float, Cq: float):
    """Properties of a transmon qubit.

    Calculate LJ, EJ, EC, wq, eps from Ic,Cq.

    Args:
        Ic (float): Junction Ic (in A)
        Cq (float): Junction capacitance (in F)

    Returns:
        tuple: [LJ, EJ, Zqp, EC, wq, wq0, eps1] -- Inductance
    """

    LJ = phi0 / Ic
    EJ = phi0**2 / LJ / hbar
    Zqp = np.sqrt(LJ / Cq)
    EC = e**2 / 2 / Cq / hbar
    wq0 = 1 / np.sqrt(LJ * Cq)
    wq = 1 / np.sqrt(LJ * Cq) - EC

    # charge dispersion
    eps1 = EC * 2**9 * (np.sqrt(2/np.pi)) * \
        (EJ/2/EC)**(1.25) * np.exp(-np.sqrt(8*EJ/EC))

    return LJ, EJ, Zqp, EC, wq, wq0, eps1


# TODO: Move to a more generic file


def chi(g: float, wr: float, w01: float, w12: float):
    r"""
    Calculate the dispersive shift $\chi$, where $2*\chi$ is
    the `|0> --> |1>` splitting).

    Accounts for push on the i-th transmon level due to the j-th transmon level,
    mediated by cavity.

    All args need to be in the same units.

    Args:
        g (float): Qubit-cavity linear coupling.
        wr (float): Frequency of resonator.
        w01 (float): Qubit 01 transition frequency
        w12 (float): Qubit 12 transition frequency

    Returns:
        float: Calculated chi value
    """

    # Push on the i-th transmon level due to the j-th transmon level
    # mediated by cavity

    # shift of the zero state
    # In this case i=0 and j=1.
    chibus_0 = -2 * g**2 * w01 / (w01**2 - wr**2)  # Koch Eq. (3.10)

    # shift of the 1 state
    # the g of the levels scales rougly as sqrt(n), so the 2 for the 2 / (w12 - wr)
    chibus_1 = g**2 * (1 / (w01 - wr) - 2 / (w12 - wr) + 1 / (w01 + wr) - 2 /
                       (w12 + wr))

    return (chibus_1 - chibus_0) / 2  # Koch Eq. (3.9)


def extract_transmon_coupled_Noscillator(capMatrix,
                                         Ic: float,
                                         CJ: float,
                                         N: int,
                                         fb: List[float],
                                         fr: float,
                                         res_L4_corr: float = None,
                                         g_scale: float = 1.0,
                                         print_info: bool = False):
    """Primary analysis function called by the user for lumped-element mode
    (LOM) analysis. Uses a Maxwell capacitance matrix generated by a capacitive
    extractor, such as Q3D. Additionally takes more values to determine many
    parameters of the Hamiltonian of the system.

    The capMatrix can be imported using `readin_q3d_matrix`

    Calculate the χ The full dispersive splitting using analytical
    approximations, i.e., return the `|0> --> |1>` splitting.

    Args:
        capMatrix (np.ndarray): Order of the capacitance matrix must be
          bus1...busN-1, ground, Qubit_pad1, Qubit_pad2, readout. (in F)
          If not in the correct order, use df_reorder_matrix_basis() to put
          it in the proper order. It is advised that the user follow a naming scheme
          in QiskitMetal or Q3D which results in the necessary order by default (eg. alphabetical)
        Ic (float): Junction Ic (in A)
        Cj (float): Junction capacitance (in F)
        N (int): Coupling pads (1 readout, N-1 bus)
        fb (float): Coupling bus and readout frequencies (in GHz).
                    fb can be a list with the order the order they appear in the capMatrix.
        fr (float): Coupling bus and readout frequencies (in GHz).
                    fr can be a list with the order the order they appear in the capMatrix.
        res_L4_corr (list): Correction factor is the resonators are L/4
          if none it ignores, otherwise this is a list of length N
          in the form [1,0,1,...].  Defaults to None.
        g_scale (float): Scale factor
        print_info (bool): Print transmon and coupling properties. Defaults to False.

    Returns:
        dict: `ham_dict` is a dictionary of all the calculated values

    Raises:
        ValueError: If N is not positive
        ValueError: If the capacitance matrix is the wrong size
    """

    # Error checks
    if N < 0:
        raise ValueError('N must positive')
    if len(capMatrix) != (N + 3):
        raise ValueError('Capacitance matrix is not the right size')

    # make list of angular frequencies of resonators
    wr = np.zeros(N)  # angular freq of resonators (GHz-rad)
    for ii in range(N):
        if ii == 0:  # readout resonator
            wr[ii] = 2 * np.pi * fr * 1e9
        else:
            if isinstance(fb, (int, float)):  # just a single one
                wr[ii] = 2 * np.pi * fb * 1e9
            else:
                wr[ii] = 2 * np.pi * fb[ii - 1] * 1e9  # offset index by

    ########################################################
    #### Transmission line properties

    # Initial values
    Zbus = 50
    Cr = 0.5 * np.pi / (wr * Zbus)
    Lr = 1 / wr**2 / Cr

    # L/4 resonators have an effective capacitance that is
    # half the L/2 case (and likewise the Lr is twice) at the
    # same resonance frequency
    # DCM (I think from numerics)
    if not res_L4_corr is None:
        for i in range(len(res_L4_corr)):
            if res_L4_corr[i]:
                Cr[i] /= 2.0
                Lr[i] *= 2.0

    ########################################################
    # Capacitance matrix parsing

    ground_index = max([0, N - 1])
    qubit_index = [ground_index + 1, ground_index + 2]
    bus_index = np.zeros(N, dtype=int)
    for ii in range(N):
        if ii == 0:
            bus_index[ii] = len(capMatrix) - 1
        else:
            bus_index[ii] = ii - 1

    # Cg list of qubit pads to ground
    Cg = [
        -capMatrix[qubit_index[0], ground_index],
        -capMatrix[qubit_index[1], ground_index]
    ]

    # Cs qubit pads to each other
    Cs = -capMatrix[qubit_index[0], qubit_index[1]]

    # Cbus (qubit pads to coupling pads)
    # index is ordered as [readout,bus1,...]
    Cbus = np.zeros([2, N])
    for ii in range(2):
        for jj in range(N):
            Cbus[ii, jj] = -capMatrix[qubit_index[ii], bus_index[jj]]

    # crosspad capacitance
    Cbusbus = np.zeros([N, N])
    for ii in range(N):
        for jj in range(N):
            if ii != jj:
                Cbusbus[ii, jj] = -capMatrix[bus_index[ii], bus_index[jj]]

    # sum of capacitances from each pad to ground
    # this assumes the bus couplers are at "ground"
    C1S = Cg[0] + np.sum(Cbus[
        0,
    ])
    C2S = Cg[1] + np.sum(Cbus[
        1,
    ])

    # total capacitance between pads
    tCSq = Cs + C1S * C2S / (C1S + C2S)  # Key equation

    # total capacitance of each pad to ground?
    # Note the + in the squared term below !!!
    tCSbus = np.zeros(N)
    for ii in range(N):
        tCSbus[ii] = Cr[ii] - (Cbus[0, ii]+Cbus[1, ii])**2 / \
            (C1S+C2S) + np.sum(Cbus[:, ii]) + np.sum(Cbusbus[ii, :])

    # qubit to coupling pad capacitance
    tCqbus = (C2S * Cbus[
        0,
    ] - Cbus[
        1,
    ] * C1S) / (C1S + C2S)

    # coupling pad to coupling pad capacitance
    tCqbusbus = np.zeros([N, N])
    for ii in range(N):
        for jj in range(N):
            tCqbusbus[ii, jj] = Cbusbus[ii, jj] + \
                (Cbus[0, ii]+Cbus[1, ii])*(Cbus[0, jj]+Cbus[1, jj])/(C1S+C2S)

    # voltage division ratio
    bbus = (C2S * Cbus[
        0,
    ] - Cbus[
        1,
    ] * C1S) / ((C1S + C2S) * Cs + C1S * C2S)

    # total qubit capacitance (including junction capacitance)
    Cq = tCSq + CJ

    ########################################################
    ##### Transmon qubit & bus quantum properties

    # get transmon properties given Ic and Cq
    LJ, EJ, Zqp, EC, wq, wq0, eps1 = transmon_props(Ic, Cq)

    # get numerical properties
    fq, alpha, disp, tphi_ms = levels_vs_ng_real_units(Cq / 1e-15,
                                                       Ic / 1e-9,
                                                       N=51)
    wq = 2 * np.pi * fq * 1e9

    # effective impedances of the coupling pads (?)
    Zbus = np.sqrt(Lr / tCSbus)

    # matrix of the qubit and couplings with a single ground node
    #    Cm = np.zeros([N+1,N+1])

    #    Cm = np.array([[Cq, -tCqr1, -tCqr2],
    #        [-tCqr1, tCSr1, -tCr1r2],
    #        [-tCqr2, -tCr1r2, tCSr2]])
    #
    #    Cminv = np.linalg.inv(Cm)
    #
    #    #readout g
    #    gr0 = 0.5*Cminv[0,1]/np.sqrt(Zqp*Zr1p)
    #    print('gr0 = %f [MHz]'%(gr0/2/np.pi/1e6))

    # g's from the qubit
    gqbus = 0.5 * wr * bbus * np.sqrt(Zbus / Zqp) * g_scale
    #gbus = bbus*wr*np.sqrt(Zbus)*e*(EJ/8/EC)**(1/4)/np.sqrt(hbar)
    gbus_in_MHz = gqbus / 1e6 / 2 / np.pi

    # g's between pads
    gbusbus = np.zeros([N, N])
    for ii in range(N):
        for jj in range(N):
            gbusbus[ii,
                    jj] = (0.01) * tCqbusbus[ii, jj] / (tCSbus[ii] * tCSbus[jj])

    ########################################################
    ##### Purcell, Qs, dissipative

    # guesses for the Q's
    Qreadout = 1e4
    Qcouplingbus = 1e5
    Qbus = np.zeros(N)
    for ii in range(N):
        if ii == 0:
            Qbus[ii] = Qreadout
        else:
            Qbus[ii] = Qcouplingbus

    # loss tangent
    kbus = wr / Qbus

    # purcell due to each coupling bus
    T1bus = (wr**2 - wq**2)**2 / (4 * kbus * gqbus**2 * wq**2)

    # total T1
    if N > 0:
        T1 = 1 / (np.sum(1 / T1bus))
    else:
        T1 = 100

    ########################################################
    ##### Transmon properties and final summary

    # chi's
    #d = -EC
    d = alpha * 2 * np.pi * 1e6
    Chi_in_MHz = 2 * chi(gqbus, wr, wq,
                         d + wq) / 2 / np.pi / 1e6  # Total chi in MHz

    ham_dict = {}
    ham_dict['fQ'] = wq / 2 / np.pi / 1E9
    ham_dict['EC'] = EC / 2 / np.pi / 1E6
    ham_dict['EJ'] = EJ / 2 / np.pi / 1E9
    # correction to the anharmonicity (from eq 52 of solgun et al.)
    #ham_dict['alpha'] = ((wq0/wq)**2*ham_dict['EC'])
    ham_dict['alpha'] = alpha
    ham_dict['dispersion'] = disp / 1e3
    ham_dict['gbus'] = gbus_in_MHz
    ham_dict['chi_in_MHz'] = Chi_in_MHz

    if print_info:
        print(qubit_index, bus_index)
        print('Predicted Values')
        print('')
        print('Transmon Properties')
        print('f_Q %f [GHz]' % ham_dict['fQ'])
        print('EC %f [MHz]' % ham_dict['EC'])
        print('EJ %f [GHz]' % ham_dict['EJ'])
        print('alpha %f [MHz]' % ham_dict['alpha'])
        print('dispersion %f [KHz]' % ham_dict['dispersion'])
        print('Lq %f [nH]' % (phi0**2 / (hbar * EJ) / 1e-9))
        print('Cq %f [fF]' % (Cq / 1e-15))
        print('T1 %f [us]' % (T1 / (1e-6)))
        print('')

        print('**Coupling Properties**')
        for ii in range(N):
            print('\ntCqbus%d %f [fF]' % (ii + 1, tCqbus[ii] / (1e-15)))
            print('gbus%d_in_MHz %f [MHz]' % (ii + 1, ham_dict['gbus'][ii]))
            print('χ_bus%d %f [MHz]' % (ii + 1, Chi_in_MHz[ii]))
            print('1/T1bus%d %f [Hz]' % (ii + 1, 1 / T1bus[ii] / (2 * np.pi)))
            print('T1bus%d %f [us]' % (ii + 1, T1bus[ii] / (1e-6)))

        print('Bus-Bus Couplings')
        for ii in range(N):
            for jj in range(ii + 1, N):
                print('gbus%d_%d %f [MHz]' % (ii + 1, jj + 1, gbusbus[ii, jj] /
                                              (2 * np.pi * 1e6)))

    return ham_dict


def levels_vs_ng_real_units(Cq, IC, N=301, do_disp=0, do_plots=0):
    """This numerically computes the exact transmon levels given C and IC as a
    function of the ng ration -- it subtracts the vacuum flucations so that
    the ground state is set to zero energy.

    Args:
        C (float): In fF
        Ic (float): In nA
        N (int): Number of charge values to use (needs to be odd)
        do_disp (int): Will print out the values
        do_plots (int): Will plot the data

    Returns:
        tuple: fqubitGHz, anharMHz, disp, tphi_ms

    Raises:
        ValueError: If the matrix is not Hermitian
    """
    C = Cq * 1e-15
    IC = IC * 1e-9
    Ec = e**2 / 2 / C

    nmax = 40
    dim = 2 * nmax + 1
    nmat = np.zeros([dim, dim])
    V = nmat
    charge = np.linspace(-1., 1., N)

    # KE
    for ii in range(dim):
        nmat[ii, ii] = ii - nmax

    # PE
    V = np.diag(-0.5 * np.ones([dim - 1]), 1)
    V = (V + np.transpose(np.conj(V)))

    varphi = hbar / 2 / e
    EJ = IC * varphi

    elvls = np.zeros([dim, N])

    for iindex in range(N):
        ng = charge[iindex]

        H = 4 * Ec * (nmat - ng * np.eye(dim))**2 + EJ * V

        if (not np.array_equal(H, np.transpose(np.conj(H)))):
            raise ValueError('Matrix is not Hermitian')

        [d, v] = np.linalg.eig(H)
        sortIX = np.argsort(d)
        sorted_d = d[sortIX]

        elvls[:, iindex] = (sorted_d - sorted_d[0])

    if do_plots:

        # plot using matplotlib (might need to clean this up)

        plt.figure()
        plt.subplot(1, 2, 1)
        plt.plot(charge,
                 elvls[
                     0,
                 ] / h / 1e9,
                 'k',
                 charge,
                 elvls[
                     1,
                 ] / h / 1e9,
                 'b',
                 charge,
                 elvls[
                     2,
                 ] / h / 1e9,
                 'r',
                 charge,
                 elvls[
                     3,
                 ] / h / 1e9,
                 'g',
                 linewidth=2)
        plt.xlabel('Gate charge, n_g [2e]')
        plt.ylabel('Energy, E_n [GHz]')
        plt.subplot(1, 2, 2)
        plt.plot(charge, (-elvls[1, :] / h + elvls[1, 0] / h) / 1e3, 'k')
        plt.xlabel('Gate charge, n_g [2e]')
        plt.ylabel('Energy [kHz], ')
        plt.show()

        plt.figure(2)
        plt.subplot(1, 2, 1)
        plt.plot(
            charge, 1000 * (elvls[
                2,
            ] / h / 1e9 - elvls[
                0,
            ] / h / 1e9 - 2 * elvls[
                1,
            ] / h / 1e9 - elvls[
                0,
            ] / h / 1e9), charge, -charge * 0 - 1000 * Ec / h / 1e9)
        plt.xlabel('Gate charge, n_g [2e]')
        plt.ylabel('delta [MHZ] green theory, blue numerics ')
        plt.subplot(1, 2, 2)
        plt.plot(charge, elvls[
            1,
        ] / h / 1e9 - elvls[
            0,
        ] / h / 1e9, charge, charge * 0 + (np.sqrt(8 * EJ * Ec) - Ec) / h / 1e9)
        plt.xlabel('Gate charge, n_g [2e]')
        plt.ylabel('F01 [GHZ] green theory, blue numerics ')
        plt.show()

    fqubitGHz = np.mean(elvls[
        1,
    ] / h / 1e9)
    anharMHz = np.mean(1000 * (elvls[
        2,
    ] / h / 1e9 - elvls[
        0,
    ] / h / 1e9 - 2 * elvls[
        1,
    ] / h / 1e9 - elvls[
        0,
    ] / h / 1e9))

    disp = np.max(-elvls[
        1,
    ] / h + elvls[1, 0] / h)
    tphi_ms = 2 / (2 * np.pi * disp * np.pi * 1e-4 * 1e-3)

    if do_disp:
        print('Mean Frequency %f [GHz]' % fqubitGHz)
        print('Anharmonicity %f [MHz]' % anharMHz)
        print('EC %f [GHz]' % (Ec / h / 1e9))
        print('Charge Dispersion %f [kHz]' % (disp / 1e3))
        print('Dephasing Time %f [ms]' % tphi_ms)

    return fqubitGHz, anharMHz, disp, tphi_ms


def get_C_and_Ic(Cin_est, Icin_est, f01, f02on2):
    """Get the capacitance and critical current for a transmon of a certain
    frequency and anharmonicity.

    Args:
        Cin_est (float): Initial guess for capacitance (in fF)
        Icin_est (float): Initial guess for critical current (in nA)
        f01 (float): Transmon frequency (in GHz)
        f02on2 (float): 02/2 frequency (in GHz)

    Returns:
        tuple: [C,Ic] from levels_vs_ng_real_units that gives the
        specified frequency and anharmonicity
    """

    xrr = opt.minimize(lambda x: cos_to_mega_and_delta(x[0], x[1], f01, f02on2),
                       [Cin_est, Icin_est],
                       tol=1e-4,
                       options={
                           'maxiter': 100,
                           'disp': True
                       })

    return xrr.x


########################################################################
# Utility functions for reporting and loading - Zlatko


# Cost function for calculating C and IC
# given a C and IC calculate f and f02/2 from 'levels_vs_ng_real_units'
# and least square with measured f01,f02on2
def cos_to_mega_and_delta(Cin, ICin, f01, f02on2):
    """Cost function for calculating C and IC given a C and IC calculate f and
    f02/2 from 'levels_vs_ng_real_units' and least square with measured
    f01,f02on2.

    Args:
        Cin (float): Cin
        ICin (float): ICin
        f01 (float): f01
        f02on2 (float): f02on2

    Returns:
        float: Calculated value
    """
    fqubitGHz, anharMHz, disp, tphi_ms = levels_vs_ng_real_units(Cin,
                                                                 ICin,
                                                                 N=51)

    return ((fqubitGHz - f01)**2 +
            (fqubitGHz + anharMHz / 2. / 1e3 - f02on2)**2)**0.5


# TODO: Move to a more generic file


def chargeline_T1(Ccharge, Cq, f01):
    """Calculate the charge line `T1`.

    Args:
        Cchare (float): Ccharge
        Cq (float): Cq
        f01 (float): f01

    Returns:
        float: Calculated chargeline T1
    """

    return Cq / (Ccharge**2 * 50. * (2 * np.pi * f01)**2)


# TODO: Move to a more generic file


def readin_q3d_matrix(path: str, delim_whitespace=True):
    """Read in the txt file created from q3d export as CSV and output the
    capacitance matrix file exported by Ansys Q3D.

    When exporting pick "save as type: data table"

    Args:
        path (str): Path to file

    Returns:
        tuple: df_cmat, units, design_variation, df_cond

    Example file:

    ::

        DesignVariation:$BBoxL='650um' $boxH='750um' $boxL='2mm' $QubitGap='30um' $QubitH='90um' $QubitL='450um' Lj_1='13nH'
        Setup1:LastAdaptive
        Problem Type:C
        C Units:farad, G Units:mSie
        Reduce Matrix:Original
        Frequency: 5.5E+09 Hz

        Capacitance Matrix
            ground_plane        Q1_bus_Q0_connector_pad Q1_bus_Q2_connector_pad Q1_pad_bot      Q1_pad_top1     Q1_readout_connector_pad
        ground_plane    2.8829E-13      -3.254E-14      -3.1978E-14     -4.0063E-14     -4.3842E-14     -3.0053E-14
        Q1_bus_Q0_connector_pad -3.254E-14      4.7257E-14      -2.2765E-16     -1.269E-14      -1.3351E-15     -1.451E-16
        Q1_bus_Q2_connector_pad -3.1978E-14     -2.2765E-16     4.5327E-14      -1.218E-15      -1.1552E-14     -5.0414E-17
        Q1_pad_bot      -4.0063E-14     -1.269E-14      -1.218E-15      9.5831E-14      -3.2415E-14     -8.3665E-15
        Q1_pad_top1     -4.3842E-14     -1.3351E-15     -1.1552E-14     -3.2415E-14     9.132E-14       -1.0199E-15
        Q1_readout_connector_pad        -3.0053E-14     -1.451E-16      -5.0414E-17     -8.3665E-15     -1.0199E-15     3.9884E-14

        Conductance Matrix
            ground_plane        Q1_bus_Q0_connector_pad Q1_bus_Q2_connector_pad Q1_pad_bot      Q1_pad_top1     Q1_readout_connector_pad
        ground_plane    0       0       0       0       0       0
        Q1_bus_Q0_connector_pad 0       0       0       0       0       0
        Q1_bus_Q2_connector_pad 0       0       0       0       0       0
        Q1_pad_bot      0       0       0       0       0       0
        Q1_pad_top1     0       0       0       0       0       0
        Q1_readout_connector_pad        0       0       0       0       0       0
    """

    text = Path(path).read_text()

    s1 = text.split('Capacitance Matrix')
    assert len(s1) == 2, "Copuld not split text to `Capacitance Matrix`"

    s2 = s1[1].split('Conductance Matrix')

    df_cmat = pd.read_csv(io.StringIO(s2[0].strip()),
                          delim_whitespace=delim_whitespace,
                          skipinitialspace=True,
                          index_col=0)
    if len(s2) > 1:
        df_cond = pd.read_csv(io.StringIO(s2[1].strip()),
                              delim_whitespace=delim_whitespace,
                              skipinitialspace=True,
                              index_col=0)
    else:
        df_cond = None

    if delim_whitespace == False and len(df_cmat.columns):
        df_cmat = df_cmat.drop(df_cmat.columns[-1], axis=1)

    units = re.findall(r'C Units:(.*?),', text)[0]
    design_variation = re.findall(r'DesignVariation:(.*?)\n', text)
    if len(design_variation) == 0:
        design_variation = re.findall(r'Design Variation:(.*?)\n', text)
        if design_variation:
            design_variation = design_variation[0]
        else:
            design_variation = ''
    else:
        design_variation = design_variation[0]

    return df_cmat, units, design_variation, df_cond


def readin_q3d_matrix_m(path: str) -> pd.DataFrame:
    """Read in Q3D cap matrix from a .m file exported by Ansys Q3d.

    Args:
        path (str): Path to .m file

    Returns:
        pd.DataFrame of cap matrix, with no names of columns.
    """
    text = Path(path).read_text()
    match = re.findall(r'capMatrix (.*?)]', text, re.DOTALL)
    if match:
        match = match[0].strip('= [').strip(']').strip('\n')
        dfC = pd.read_csv(io.StringIO(match),
                          skipinitialspace=True,
                          header=None)
    return dfC


# TODO: Move to a more generic file


def load_q3d_capacitance_matrix(path, user_units='fF', _disp=True):
    """Load Q3D capcitance file exported as Maxwell matrix. Do not export
    conductance. Units are read in automatically and converted to user units.

    Args:
        path (str): Path to file.
        user_units (str): Units.  Defaults to 'fF'.
        _disp (bool): whehter or not to display messages.  Defaults to True.

    Returns:
        tupe: df_cmat, user_units, design_variation, df_cond
    """
    df_cmat, Cunits, design_variation, df_cond = readin_q3d_matrix(path)

    # Unit convert
    ureg = UnitRegistry()
    q = ureg.parse_expression(Cunits).to(user_units)
    df_cmat = df_cmat * q.magnitude  # scale to user units

    # Report
    if _disp:
        print(
            "Imported capacitance matrix with UNITS: [%s] now converted to USER UNITS:[%s] \
                from file:\n\t%s" % (Cunits, user_units, path))
        df_cmat_style_print(df_cmat)

    return df_cmat, user_units, design_variation, df_cond


def lumped_oscillator_from_path(path: str, Lj_nH: float, Cj_fF: float, N: int,
                                fr: Union[list, float],
                                fb: Union[list, float]) -> pd.DataFrame:
    """Obtain a single result dataframe from a Q3D capacitance file pointed to by path.
    Similar to member method lumped_oscillator_vs_passes() of QQ3DRenderer but for a user
    provided capacitance matrix file.

    Args:
        path (str): Path to file.
        Lj_nH (float): Junction inductance (in nH)
        Cj_fF (float): Junction capacitance (in fF)
        N (int): Coupling pads (1 readout, N - 1 bus)
        fr (Union[list, float]):Readout frequencies (in GHz). fr can be a list with the order
            they appear in the capMatrix.
        fb (Union[list, float]): Coupling bus frequencies (in GHz). fb can be a list with the order
            they appear in the capMatrix.
        g_scale (float, optional): Scale factor. Defaults to 1..

    Returns:
        dict: A single dataframe corresponding to a single capacitance matrix
    """
    ureg = UnitRegistry()
    IC_Amps = Convert.Ic_from_Lj(Lj_nH, 'nH', 'A')
    CJ = ureg(f'{Cj_fF} fF').to('farad').magnitude
    fr = ureg(f'{fr} GHz').to('GHz').magnitude
    fb = [ureg(f'{freq} GHz').to('GHz').magnitude for freq in fb]

    df_cmat, user_units, _, _, = load_q3d_capacitance_matrix(path)
    c_units = ureg(user_units).to('farads').magnitude

    RES = extract_transmon_coupled_Noscillator(df_cmat.values * c_units,
                                               IC_Amps,
                                               CJ,
                                               N,
                                               fb,
                                               fr,
                                               g_scale=1,
                                               print_info=True)

    RES = pd.DataFrame([RES])
    RES['χr MHz'] = abs(RES['chi_in_MHz'].apply(lambda x: x[0]))
    RES['gr MHz'] = abs(RES['gbus'].apply(lambda x: x[0]))
    return RES


def df_cmat_style_print(df_cmat: pd.DataFrame):
    """Display the dataframe in the cmat style.

    Args:
        df_cmat (dataframe): Dataframe to display
    """
    from IPython.display import display
    display(df_cmat.style.format("{:.2f}").bar(color='#5fba7d', width=100))


########################################################################
# Utility functions - Zlatko


def move_index_to(i_from: List[int], i_to: List[int], len_):
    """Utility function to swap index.

    Args:
        i_from (int): Data frame to swap index
        i_to (int): Data frame to index
        len_ (int): Length of array

    Returns:
        list: list of indecies, such as  array([1, 2, 3, 4, 0, 5])
    """
    idxs = np.arange(0, len_)
    idxs = np.delete(idxs, i_from)
    return np.insert(idxs, i_to, i_from)


# TODO: Move to a more generic file


def df_reorder_matrix_basis(df, i_from, i_to):
    """Data frame handle reording of matrix basis.

    Args:
        df (DataFrame): Data frame to swap
        i_from (int): Index to move from
        i_to (int): Index to move to

    Returns:
        DataFrame: The changed index DataFrame
    """
    arr = df.values
    idx = move_index_to(i_from, i_to, len(arr))
    arr = arr[np.ix_(idx, idx)]
    # Maybe make copy
    return pd.DataFrame(arr, columns=df.columns[idx], index=df.index[idx])



import numpy as np

MHzRad = 2 * np.pi * 1e6
GHzRad = 2 * np.pi * 1e9

NANO = 1e-9
FEMTO = 1e-15
ONE_OVER_FEMTO = 1e15

e = 1.60217657e-19  # electron charge
h = 6.62606957e-34  # Plank's
hbar = 1.0545718E-34  # Plank's reduced
phinot = 2.067 * 1E-15  # magnetic flux quantum
phi0 = phinot / (2 * np.pi)  # reduced magnetic flux quantum


# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from typing import TYPE_CHECKING, List

from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QModelIndex, Qt, QTimer
from PySide6.QtGui import QContextMenuEvent
from PySide6.QtWidgets import (QAbstractItemView, QInputDialog, QLabel,
                               QLineEdit, QMenu, QMessageBox, QTableView,
                               QVBoxLayout)

from ...utility._handle_qt_messages import slot_catch_error
from ..bases.QWidget_PlaceholderText import QWidget_PlaceholderText

if TYPE_CHECKING:
    from ...main_window import MetalGUI
    from .table_model_all_components import QTableModel_AllComponents


class QTableView_AllComponents(QTableView, QWidget_PlaceholderText):
    """Design components.

    The class extends the `QTableView` and `QWidget_PlaceholderText` classes.

    Table model that shows the summary of the components of a design in a table
    with their names, classes, and modules

    Access:
        table = gui.ui.tableComponents
    """

    def __init__(self, parent: QtWidgets.QWidget):
        """
        Args:
            parent (QWidget): Parent widget
        """
        QTableView.__init__(self, parent)
        QWidget_PlaceholderText.__init__(
            self,
            "No QComponents to show.\n\nCreate components from the QLibrary.",
            self)

        # Connect signals
        self.clicked.connect(self.viewClicked)
        self.doubleClicked.connect(self.doDoubleClicked)

        # Configure selection behavior
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setSelectionBehavior(QTableView.SelectRows)

        # Apply styling
        QTimer.singleShot(100, self.style2)

    def style2(self):
        """Style the widget."""
        # Do in the ui file
        self.horizontalHeader().hide()
        self.verticalHeader().show()

        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)

        selection_model = self.selectionModel()
        selection_model.selectionChanged.connect(self.selection_changed)

    @property
    def design(self):
        """Returns the design."""
        return self.model().sourceModel().design

    @property
    def logger(self):
        """Returns the logger."""
        return self.model().sourceModel().logger

    @property
    def gui(self) -> 'MetalGUI':
        """Returns the GUI."""
        return self.model().sourceModel().gui

    # @slot_catch_error
    def contextMenuEvent(self, event: QContextMenuEvent):
        """This event handler, for event event, can be reimplemented in a
        subclass to receive widget context menu events.

        The handler is called when the widget's contextMenuPolicy
        is Qt::DefaultContextMenu.

        The default implementation ignores the context event.
        See the QContextMenuEvent documentation for more details.

        Args:
            event (QContextMenuEvent): The event
        """
        self._event = event  # debug

        # TODO: Should we work based on slection or what we have clicked at the moment?
        self.menu = QMenu(self)
        self.menu._d = self.menu.addAction("Delete")
        self.menu._r = self.menu.addAction("Rename")

        self.menu._d.triggered.connect(lambda: self.do_menu_delete(event))
        self.menu._r.triggered.connect(lambda: self.do_menu_rename(event))

        self.menu._action = self.menu.exec_(self.mapToGlobal(event.pos()))

    def get_name_from_event(self, event):
        """Get the event name.

        Args:
            event (QContextMenuEvent): The event

        Returns:
            tuple: Event name, Row the event is on
        """
        # get the selected row and column
        row = self.rowAt(event.pos().y())
        #col = self.columnAt(event.pos().x())

        model = self.model()
        index = model.index(row, 0)  # get the name
        name = model.data(index)
        return name, row

    # @slot_catch_error
    def do_menu_delete(self, event):
        """Called when the user clicks the context menu delete.

        Args:
            event (QContextMenuEvent): The event
        """
        name, row = self.get_name_from_event(event)

        if row > -1:
            ret = QMessageBox.question(
                self, '', f"Are you sure you want to delete component {name}",
                QMessageBox.Yes | QMessageBox.No)
            if ret == QMessageBox.Yes:
                self._do_delete(name)

    def _do_delete(self, name: str):
        """Do delte a component by name.

        Args:
            name (str): Name of the component to delete
        """
        self.logger.info(f'Deleting {name}')
        self.design.delete_component(name)
        # replot
        self.gui.plot_win.replot()

    def do_menu_rename(self, event):
        """Called when the user clicks the context menu rename.

        Args:
            event (QContextMenuEvent): The event
        """
        name, row = self.get_name_from_event(event)

        self.do_menu_rename_helper(name, row)

    def do_menu_rename_helper(self, name: str, row: int):
        """
        Allows users to rename a created QComponent
        Args:
            name (str): Old name for QComponent
            row (int): Row of QComponent in Model

        """
        if row > -1:
            text, okPressed = QInputDialog.getText(self,
                                                   f"Rename component {name}",
                                                   f"Rename {name} to:",
                                                   QLineEdit.Normal, "")
            if okPressed and text != '':
                self.logger.info(f'Renaming {name} to {text}')
                comp_id = self.design.components[name].id
                self.design.rename_component(comp_id, text)

    def viewClicked(self, index: QModelIndex):
        """Select a component and set it in the compoient widget when you left
        click.

        In the init, we had to connect with self.clicked.connect(self.viewClicked)

        Args:
            index (QModelIndex): The index
        """
        if self.gui is None or not index.isValid():
            return

        # get the component name
        # model = clickedIndex.model()  # type: QTableModel_AllComponents
        name = index.sibling(index.row(), 0).data()
        self.logger.debug(f'Selected component {name}')

        gui = self.gui
        gui.edit_component(name)
        gui.ui.dockComponent.show()
        gui.ui.dockComponent.raise_()

    def doDoubleClicked(self, index: QModelIndex):
        """
        SIGNAL: doubleClicked

        This signal is emitted when a mouse button is double-clicked.
        The item the mouse was double-clicked on is specified by index.
        The signal is only emitted when the index is valid.

        Note that single click will also get called.

        Args:
            index (QModelIndex): The index
        """
        if self.gui is None or not index.isValid():
            return

        # name of component
        name = index.sibling(index.row(), 0).data()
        self.gui.canvas.zoom_on_components([name])
        # self.logger.info(f'Double clicked component {name}')

    def rows_to_names(self, rows: List[int]) -> List[str]:
        """Based on user highlighting  rows of components in GUI, return the
        name of components.

        Args:
            rows (List[int]): User highlighted rows.

        Returns:
            List[str]: List of components that user highlighted.
        """

        def get_name(row):
            return self.model().data(self.model().index(row, 0))  # get the name

        selected_names = [get_name(row) for row in rows]
        return selected_names

    def selection_changed(self, *args):
        """Update by highlighting, the rows which the user selected."""
        rows = set([idx.row() for idx in self.selectedIndexes()])
        selected_names = self.rows_to_names(rows)

        self.logger.debug(f'Highlighting {selected_names}')
        self.gui.highlight_components(selected_names)

    def delete_selected_rows(self, *args):
        """
        Deletes selected rows of QComponents

        Args:
            *args: Allows function to be a slot
            even for signals that pass in args

        """
        index_list = self.selectedIndexes()
        model = self.model()

        index_dict = {}
        name_set = set()
        for ind in index_list:
            if ind.row() not in index_dict:
                index_dict[ind.row()] = ind
                name_index = model.index(ind.row(), 0)
                name = model.data(name_index)
                name_set.add(name)

        for name in name_set:
            if name is not None:
                ret = QMessageBox.question(
                    self, '',
                    f"Are you sure you want to delete component {name}",
                    QMessageBox.Yes | QMessageBox.No)
                if ret == QMessageBox.Yes:
                    self._do_delete(name)

    def rename_row(self, *args):
        """
        Rename single component
        Args:
            *args: Allows function to be a slot
            even for signals that pass in args

        """

        index_list = self.selectedIndexes()
        if len(index_list) == 0:
            return

        index = index_list[0]
        for ind in index_list:
            if ind.row() != index.row():
                return
        model = self.model()
        name_index = model.index(index.row(), 0)
        name = model.data(name_index)
        self.do_menu_rename_helper(name, index.row())

    def name_of_selected_qcomponent(self):
        """
        Returns names of selected qcomponents
        Returns:
            list(str): names of selected qcomponents

        """
        model = self.model()
        index_list = self.selectedIndexes()
        name_set = set()
        if len(index_list) == 0:
            return

        for ind in index_list:
            name_ind = model.index(ind.row(), 0)
            name = model.data(name_ind)
            name_set.add(name)

        return list(name_set)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Tree model for component options menu."""

from typing import TYPE_CHECKING
from PySide6.QtCore import QModelIndex, Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QTreeView, QWidget
from ..bases.dict_tree_base import LeafNode, BranchNode, QTreeModel_Base, parse_param_from_str

if TYPE_CHECKING:
    from ...main_window import MetalGUI


class QTreeModel_Options(QTreeModel_Base):
    """Tree model for component options menu. Overrides rowCount method to
    include placeholder text, and data method to include 3rd column for parsed
    values.

    Args:
        QTreeModel_Base (QAbstractItemModel): Base class for nested dicts
    """

    def __init__(self, parent: QWidget, gui: 'MetalGUI', view: QTreeView):
        """Editable table with drop-down rows for component options. Organized
        as a tree model where child nodes are more specific properties of a
        given parent node.

        Args:
            parent (QWidget): The parent widget
            gui (MetalGUI): The main user interface
            view (QTreeView): View corresponding to a tree structure
        """
        super().__init__(parent=parent, gui=gui, view=view, child='component')
        self.headers = ['Name', 'Value',
                        'Parsed value']  # 3 columns instead of 2

    @property
    def data_dict(self) -> dict:
        """Return a reference to the component options (nested) dictionary."""
        return self.component.options

    def rowCount(self, parent: QModelIndex):
        """Get the number of rows.

        Args:
            parent (QModelIndex): The parent

        Returns:
            int: The number of rows
        """
        # If there is no component, then show placeholder text
        if self.component is None:
            if self._view:
                self._view.show_placeholder_text()
            return 0

        else:
            # We have a component selected
            # if we have the view of the model linked (we should)
            # hide placeholder text
            if self._view:
                self._view.hide_placeholder_text()

            ###
            # Count number of child nodes belonging to the parent.
            node = self.nodeFromIndex(parent)
            if (node is None) or isinstance(node, LeafNode):
                return 0
            return len(node)

    def data(self, index: QModelIndex, role: Qt.ItemDataRole = Qt.DisplayRole):
        """Gets the node data.

        Args:
            index (QModelIndex): Index to get data for
            role (Qt.ItemDataRole): The role..  Defaults to Qt.DisplayRole.

        Returns:
            object: fetched data
        """
        if not index.isValid():
            return None

        if self.component is None:
            return None

        # The data in a form suitable for editing in an editor. (QString)
        if role == Qt.EditRole:
            return self.data(index, Qt.DisplayRole)

        # Bold the first
        if (role == Qt.FontRole) and (index.column() == 0):
            font = QFont()
            font.setBold(True)
            return font

        if role == Qt.TextAlignmentRole:
            return int(Qt.AlignTop | Qt.AlignLeft)

        if role == Qt.DisplayRole:
            node = self.nodeFromIndex(index)
            if node:
                # the first column is either a leaf key or a branch
                # the second column is always a leaf value or for a branch is ''.
                if isinstance(node, BranchNode):
                    # Handle a branch (which is a nested subdictionary, which can be expanded)
                    if index.column() == 0:
                        return node.name
                    return ''
                # We have a leaf
                elif index.column() == 0:
                    return str(node.label)  # key
                elif index.column() == 1:
                    return str(node.value)  # value
                elif index.column() == 2:
                    # TODO: If the parser fails, this can throw an error
                    return str(self.design.parse_value(node.value))
                else:
                    return None

        return None



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from PySide6.QtWidgets import (QAbstractItemView, QMainWindow, QMessageBox)

from .list_model_base import DynamicList
from .renderer_hfss_model import RendererHFSS_Model
from .renderer_hfss_ui import Ui_MainWindow
from .endcap_hfss_gui import EndcapHFSSWidget


class RendererHFSSWidget(QMainWindow):
    """Contains methods associated with Ansys HFSS Renderer button."""

    def __init__(self, parent: 'QMainWindow', gui: 'MetalGUI'):
        """Get access to design, which has the components. Then set up the
        model and view.

        Args:
            parent (QMainWindow): The parent window
            gui (MetalGUI): The metal GUI
        """
        super().__init__(parent)

        # Access design and Metal GUI:
        self._gui = gui

        # Use UI template from Qt Designer:
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        # Set up models for component list and options tree:
        self.listView = self.ui.listView
        self.list_model = DynamicList(self.design)
        self.listView.setModel(self.list_model)

        self.tree_model = RendererHFSS_Model(self, gui, self.ui.treeView)
        self.ui.treeView.setModel(self.tree_model)
        self.ui.treeView.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.ui.treeView.setHorizontalScrollMode(
            QAbstractItemView.ScrollPerPixel)

    def set_design(self, new_design: 'QDesign'):
        """Swaps out reference to design, which changes the reference to the
        dictionary.

        Args:
            new_design (QDesign): The design
        """
        self.list_model.update_src(self.design)

    @property
    def design(self):
        """Returns the design."""
        return self._gui.design

    def refresh(self):
        """Refreshes list of components."""
        self.list_model.populate_list()

    def select_all(self):
        """Marks all components for export."""
        self.list_model.select_all()

    def deselect_all(self):
        """Empties list of components to export."""
        self.list_model.deselect_all()

    def get_checked(self) -> list:
        """Gets list of all selected components to export.

        Returns:
            list: List of selected components
        """
        return self.list_model.get_checked()

    def choose_checked_components(self):
        """Sends checked components to the next window that allows the user to
        select which unconnected pins are to be open."""
        components_to_render = self.get_checked()
        if components_to_render:
            txt = self.ui.solnComboBox.currentText(
            )  # solution type - eigenmode or driven modal
            self.endcap_hfss_window = EndcapHFSSWidget(self, self._gui,
                                                       components_to_render,
                                                       txt)
            self.endcap_hfss_window.show()
            self.close()
        else:
            QMessageBox.warning(self, "Error",
                                "Please select at least one component.")



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Child of transmon cross, adds a flux line (galvanic T) to the arm with the
DC SQUID."""
# pylint: disable=invalid-name
# Modification of Transmon Pocket Object to include a charge line (would be better to just make as a child)

import numpy as np
from qiskit_metal import draw, Dict
from qiskit_metal.qlibrary.qubits.transmon_cross import TransmonCross


class TransmonCrossFL(TransmonCross):  # pylint: disable=invalid-name
    """The base `TransmonCrossFL` class.

    Inherits `TransmonCross` class

    Description:
        Simple Metal Transmon Cross object. Creates the X cross-shaped island,
        the "junction" on the south end, and up to 3 connectors on the remaining arms
        (claw or gap).

        'claw_width' and 'claw_gap' define the width/gap of the CPW line that
        makes up the connector. Note, DC SQUID currently represented by single
        inductance sheet

        Add connectors to it using the `connection_pads` dictonary. See BaseQubit for more
        information.

        Flux line is added by default to the 'south' arm where the DC SQUID is located,
        default is a symmetric T style

        Default Options:
        Convention: Values (unless noted) are strings with units included, (e.g., '30um')

        * make_fl -         (Boolean) If True, adds a flux line
        * t_top -           length of the flux line for mutual inductance to the SQUID
        * t_inductive_gap - amount of metallization between the flux line and SQUID
        * t_offset -        degree by which the tail of the T is offset from the center
        * t_width -         width of the flux line's transmission line center trace
        * t_gap -           dielectric gap of the flux line's transmission line

    .. image::
        TransmonCrossFL.png

    .. meta::
        :description: Transmon Cross Flux Line

    """

    component_metadata = Dict(short_name='Q',
                              _qgeometry_table_poly='True',
                              _qgeometry_table_path='True')
    """Component metadata"""

    default_options = Dict(make_fl=True,
                           fl_options=Dict(t_top='15um',
                                           t_offset='0um',
                                           t_inductive_gap='3um',
                                           t_width='5um',
                                           t_gap='3um'))
    """Default drawing options"""

    TOOLTIP = """The base `TransmonCrossFL` class."""

    def make(self):
        """Define the way the options are turned into QGeometry."""
        super().make()

        if self.options.make_fl == True:
            self.make_flux_line()


#####################################################################

    def make_flux_line(self):
        """Creates the charge line if the user has charge line option to
        TRUE."""

        # Grab option values
        pf = self.p.fl_options
        p = self.p
        #Make the T flux line
        h_line = draw.LineString([(-pf.t_top / 2, 0), (pf.t_top / 2, 0)])
        v_line = draw.LineString([(pf.t_offset, 0), (pf.t_offset, -0.03)])

        parts = [h_line, v_line]

        # Move the flux line down to the SQUID
        parts = draw.translate(
            parts, 0, -(p.cross_length + p.cross_gap + pf.t_inductive_gap +
                        pf.t_width / 2 + pf.t_gap))

        # Rotate and translate based on crossmon location
        parts = draw.rotate(parts, p.orientation, origin=(0, 0))
        parts = draw.translate(parts, p.pos_x, p.pos_y)

        [h_line, v_line] = parts

        # Adding to qgeometry table
        self.add_qgeometry('path', {
            'h_line': h_line,
            'v_line': v_line
        },
                           width=pf.t_width,
                           layer=p.layer)

        self.add_qgeometry('path', {
            'h_line_sub': h_line,
            'v_line_sub': v_line
        },
                           width=pf.t_width + 2 * pf.t_gap,
                           subtract=True,
                           layer=p.layer)

        # Generating pin
        pin_line = v_line.coords
        self.add_pin("flux_line",
                     points=pin_line,
                     width=pf.t_width,
                     gap=pf.t_gap,
                     input_as_norm=True)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable-msg=unnecessary-pass
# pylint: disable-msg=too-many-public-methods
"""Qiskit Metal unit tests components functionality."""

import unittest

from qiskit_metal.qlibrary.core import qubit, qroute
from qiskit_metal.qlibrary.sample_shapes import circle_caterpillar
from qiskit_metal.qlibrary.sample_shapes import circle_raster
from qiskit_metal.qlibrary.sample_shapes import rectangle
from qiskit_metal.qlibrary.sample_shapes import rectangle_hollow
from qiskit_metal.qlibrary.sample_shapes import n_gon
from qiskit_metal.qlibrary.sample_shapes import n_square_spiral
from qiskit_metal.qlibrary.lumped.cap_n_interdigital import CapNInterdigital
from qiskit_metal.qlibrary.couplers.coupled_line_tee import CoupledLineTee
from qiskit_metal.qlibrary.couplers.cap_n_interdigital_tee import CapNInterdigitalTee
from qiskit_metal.qlibrary.couplers.line_tee import LineTee
from qiskit_metal.qlibrary.terminations import open_to_ground
from qiskit_metal.qlibrary.terminations import short_to_ground
from qiskit_metal.qlibrary.tlines.anchored_path import RouteAnchors
from qiskit_metal.qlibrary.tlines import straight_path
from qiskit_metal.qlibrary.tlines import meandered
from qiskit_metal.qlibrary.tlines.mixed_path import RouteMixed
from qiskit_metal.qlibrary.tlines.pathfinder import RoutePathfinder
from qiskit_metal import designs
from qiskit_metal.qlibrary.terminations.launchpad_wb import LaunchpadWirebond
from qiskit_metal.qlibrary.terminations.launchpad_wb_coupled import LaunchpadWirebondCoupled
from qiskit_metal.qlibrary.lumped.cap_3_interdigital import Cap3Interdigital
from qiskit_metal.qlibrary.qubits import star_qubit
from qiskit_metal.qlibrary.qubits.JJ_Dolan import jj_dolan
from qiskit_metal.qlibrary.qubits.JJ_Manhattan import jj_manhattan
from qiskit_metal.qlibrary.qubits import transmon_concentric
from qiskit_metal.qlibrary.qubits import transmon_cross
from qiskit_metal.qlibrary.qubits.transmon_cross_fl import TransmonCrossFL
from qiskit_metal.qlibrary.qubits import transmon_pocket
from qiskit_metal.qlibrary.qubits import transmon_pocket_cl
from qiskit_metal.qlibrary.qubits.transmon_pocket_6 import TransmonPocket6
from qiskit_metal.qlibrary.qubits.transmon_pocket_teeth import TransmonPocketTeeth
from qiskit_metal.qlibrary.qubits.SQUID_loop import SQUID_LOOP
from qiskit_metal.qlibrary.couplers.tunable_coupler_01 import TunableCoupler01
from qiskit_metal.qlibrary import _template
from qiskit_metal.tests.assertions import AssertionsMixin

#pylint: disable-msg=line-too-long
from qiskit_metal.qlibrary.lumped.resonator_coil_rect import ResonatorCoilRect


class TestComponentOptions(unittest.TestCase, AssertionsMixin):
    """Unit test class."""

    def setUp(self):
        """Setup unit test."""
        pass

    def tearDown(self):
        """Tie any loose ends."""
        pass

    def test_qlibrary_circle_caterpiller_options(self):
        """Test that default options of circle_caterpiller in
        circle_caterpillar.py were not accidentally changed."""
        # Setup expected test results
        _design = designs.DesignPlanar()
        _circle_caterpillar = circle_caterpillar.CircleCaterpillar(
            _design, 'my_name')
        _options = _circle_caterpillar.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(_options), 7)
        self.assertEqual(_options['segments'], '5')
        self.assertEqual(_options['distance'], '1.2')
        self.assertEqual(_options['radius'], '300um')
        self.assertEqual(_options['resolution'], '16')
        self.assertEqual(_options['cap_style'], 'round')
        self.assertEqual(_options['subtract'], 'False')
        self.assertEqual(_options['helper'], 'False')

    def test_qlibrary_circle_raster_options(self):
        """Test that default options of circle_raster in circle_raster.py were
        not accidentally changed."""
        # Setup expected test results
        _design = designs.DesignPlanar()
        _circle_raster = circle_raster.CircleRaster(_design, 'my_name')
        _options = _circle_raster.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(_options), 5)
        self.assertEqual(_options['radius'], '300um')
        self.assertEqual(_options['resolution'], '16')
        self.assertEqual(_options['cap_style'], 'round')
        self.assertEqual(_options['subtract'], 'False')
        self.assertEqual(_options['helper'], 'False')

    def test_qlibrary_rectangle_options(self):
        """Test that default options of rectangle in rectangle.py were not
        accidentally changed."""
        # Setup expected test results
        _design = designs.DesignPlanar()
        _rectangle = rectangle.Rectangle(_design, 'my_name')
        _options = _rectangle.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(_options), 4)
        self.assertEqual(_options['width'], '500um')
        self.assertEqual(_options['height'], '300um')
        self.assertEqual(_options['subtract'], 'False')
        self.assertEqual(_options['helper'], 'False')

    def test_qlibrary_rectangle_hollow_options(self):
        """Test that default options of rectangle_hollow in rectangle_hollow.py
        were not accidentally changed."""
        # Setup expected test results
        _design = designs.DesignPlanar()
        _rectangle_hollow = rectangle_hollow.RectangleHollow(_design, 'my_name')
        _options = _rectangle_hollow.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(_options), 5)
        self.assertEqual(_options['width'], '500um')
        self.assertEqual(_options['height'], '300um')
        self.assertEqual(_options['subtract'], 'False')
        self.assertEqual(_options['helper'], 'False')

        self.assertEqual(len(_options['inner']), 5)
        self.assertEqual(_options['inner']['width'], '250um')
        self.assertEqual(_options['inner']['height'], '100um')
        self.assertEqual(_options['inner']['offset_x'], '40um')
        self.assertEqual(_options['inner']['offset_y'], '-20um')
        self.assertEqual(_options['inner']['orientation'], '15')

    def test_qlibrary_n_gon_options(self):
        """Test that default options of NGon in n_gon.py were not accidentally
        changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        my_n_gon = n_gon.NGon(design, 'my_name')
        options = my_n_gon.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 4)
        self.assertEqual(options['n'], '3')
        self.assertEqual(options['radius'], '30um')
        self.assertEqual(options['subtract'], 'False')
        self.assertEqual(options['helper'], 'False')

    def test_qlibrary_n_square_spiral_options(self):
        """Test that default options of NSquareSpiral in n_square_spiral.py
        were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        my_n_square_spiral = n_square_spiral.NSquareSpiral(design, 'my_name')
        options = my_n_square_spiral.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 6)
        self.assertEqual(options['n'], '3')
        self.assertEqual(options['width'], '1um')
        self.assertEqual(options['radius'], '40um')
        self.assertEqual(options['gap'], '4um')
        self.assertEqual(options['subtract'], 'False')
        self.assertEqual(options['helper'], 'False')

    def test_qlibrary_basequbit_options(self):
        """Test that default options of BaseQubit in qubit.py were not
        accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        my_base_qubit = qubit.BaseQubit(design, 'my_name', make=False)
        options = my_base_qubit.default_options

        # Test all elements of the results data against expected ata
        self.assertEqual(len(options), 2)
        self.assertEqual(options['connection_pads'], {})
        self.assertEqual(options['_default_connection_pads'], {})

    def test_qlibrary_open_to_ground_options(self):
        """Test that default options of OpenToGround in open_to_ground.py were
        not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        my_open_to_ground = open_to_ground.OpenToGround(design, 'my_name')
        options = my_open_to_ground.default_options

        # Test all elements of the results data against expected ata
        self.assertEqual(len(options), 3)
        self.assertEqual(options['width'], '10um')
        self.assertEqual(options['gap'], '6um')
        self.assertEqual(options['termination_gap'], '6um')

    def test_qlibrary_short_to_ground_options(self):
        """Test that default options of ShortToGround in short_to_ground.py
        where not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        my_short_to_ground = short_to_ground.ShortToGround(design, 'my_name')
        options = my_short_to_ground.default_options

        # Test all elements of the results data against expected ata
        self.assertEqual(len(options), 1)
        self.assertEqual(options['width'], '10um')

    def test_qlibrary_straight_path_options(self):
        """Test that default options of RouteStraight in straight_path.py were
        not accidentally changed."""
        # Setup expected test results
        my_straight_path = straight_path.RouteStraight
        options = my_straight_path.default_options

        # Test all elements of the results data against expected ata
        self.assertEqual(len(options), 5)
        self.assertEqual(options['fillet'], '0')
        self.assertEqual(options['total_length'], '7mm')
        self.assertEqual(options['trace_width'], 'cpw_width')

        self.assertEqual(len(options['pin_inputs']), 2)
        self.assertEqual(len(options['pin_inputs']['start_pin']), 2)
        self.assertEqual(len(options['pin_inputs']['end_pin']), 2)
        self.assertEqual(options['pin_inputs']['start_pin']['component'], '')
        self.assertEqual(options['pin_inputs']['start_pin']['pin'], '')
        self.assertEqual(options['pin_inputs']['end_pin']['component'], '')
        self.assertEqual(options['pin_inputs']['end_pin']['pin'], '')

        self.assertEqual(len(options['lead']), 4)
        self.assertEqual(options['lead']['start_straight'], '0mm')
        self.assertEqual(options['lead']['end_straight'], '0mm')
        self.assertEqual(options['lead']['start_jogged_extension'], '')
        self.assertEqual(options['lead']['end_jogged_extension'], '')

    def test_qlibrary_route_meander_options(self):
        """Test that default options of RouteMeander in meandered.py were not
        accidentally changed."""
        # Setup expected test results
        my_route_meander = meandered.RouteMeander
        options = my_route_meander.default_options

        # Test all elements of the results data against expected ata
        self.assertEqual(len(options), 3)
        self.assertEqual(options['snap'], 'true')
        self.assertEqual(options['prevent_short_edges'], 'true')

        self.assertEqual(len(options['meander']), 2)
        self.assertEqual(options['meander']['spacing'], '200um')
        self.assertEqual(options['meander']['asymmetry'], '0um')

    def test_qlibrary_route_mixed_options(self):
        """Test that default options of RouteMixed in mixed_path.py were not
        accidentally changed."""
        # Setup expected test results
        my_route_mixed = RouteMixed
        options = my_route_mixed.default_options

        # Test all elements of the results data against expected ata
        self.assertEqual(len(options), 1)
        self.assertEqual(options['between_anchors'], {})

    def test_qlibrary_my_qcomponent_options(self):
        """Test that default options in MyQComponent in _template.py were not
        accidentally changed."""
        # Setup expected test results
        my_qcomponent_local = _template.MyQComponent
        options = my_qcomponent_local.default_options

        # Test all elements of the results data against expected ata
        self.assertEqual(len(options), 2)
        self.assertEqual(options['width'], '500um')
        self.assertEqual(options['height'], '300um')

    def test_qlibrary_transmon_concentric_options(self):
        """Test that default options of transmon_concentric in
        transmon_concentric.py were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        my_transmon_concentric = transmon_concentric.TransmonConcentric(
            design, 'my_name')
        options = my_transmon_concentric.default_options

        self.assertEqual(len(options), 17)
        self.assertEqual(options['width'], '1000um')
        self.assertEqual(options['height'], '1000um')
        self.assertEqual(options['rad_o'], '170um')
        self.assertEqual(options['rad_i'], '115um')
        self.assertEqual(options['gap'], '35um')
        self.assertEqual(options['jj_w'], '10um')
        self.assertEqual(options['res_s'], '100um')
        self.assertEqual(options['res_ext'], '100um')
        self.assertEqual(options['fbl_rad'], '100um')
        self.assertEqual(options['fbl_sp'], '100um')
        self.assertEqual(options['fbl_gap'], '80um')
        self.assertEqual(options['fbl_ext'], '300um')
        self.assertEqual(options['pocket_w'], '1500um')
        self.assertEqual(options['pocket_h'], '1000um')
        self.assertEqual(options['cpw_width'], '10.0um')
        self.assertEqual(options['inductor_width'], '5.0um')

    def test_qlibrary_star_qubit_options(self):
        """Test that default options of transmon_concentric in
        star_qubit.py were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        my_star_qubit = star_qubit.StarQubit(design, 'my_name')
        options = my_star_qubit.default_options

        self.assertEqual(len(options), 18)
        self.assertEqual(options['radius'], '300um')
        self.assertEqual(options['center_radius'], '100um')
        self.assertEqual(options['gap_couplers'], '25um')
        self.assertEqual(options['gap_readout'], '10um')
        self.assertEqual(options['connector_length'], '75um')
        self.assertEqual(options['trap_offset'], '20um')
        self.assertEqual(options['junc_h'], '100um')
        self.assertEqual(options['cpw_width'], '0.01')
        self.assertEqual(options['rotation_cpl1'], '0.0')
        self.assertEqual(options['rotation_cpl2'], '72.0')
        self.assertEqual(options['rotation_rdout'], '144.0')
        self.assertEqual(options['rotation_cpl3'], '216.0')
        self.assertEqual(options['rotation_cpl4'], '288.0')
        self.assertEqual(options['number_of_connectors'], '4')
        self.assertEqual(options['resolution'], '16')
        self.assertEqual(options['cap_style'], 'round')
        self.assertEqual(options['subtract'], 'False')
        self.assertEqual(options['helper'], 'False')

    def test_qlibrary_transmon_cross_fl_options(self):
        """Test that default_options of transmon_cross_fl were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        transmon_cross_fl = TransmonCrossFL(design, 'my_name')
        options = transmon_cross_fl.default_options

        self.assertEqual(len(options), 2)
        self.assertEqual(len(options['fl_options']), 5)
        self.assertEqual(options['make_fl'], True)
        self.assertEqual(options['fl_options']['t_top'], '15um')
        self.assertEqual(options['fl_options']['t_offset'], '0um')
        self.assertEqual(options['fl_options']['t_inductive_gap'], '3um')
        self.assertEqual(options['fl_options']['t_width'], '5um')
        self.assertEqual(options['fl_options']['t_gap'], '3um')

    def test_qlibrary_transmon_pocket_6_options(self):
        """Test that default_options of transmon_pocket_6 were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        transmon_pocket_6 = TransmonPocket6(design, 'my_name')
        options = transmon_pocket_6.default_options

        self.assertEqual(len(options), 7)
        self.assertEqual(options['pad_gap'], '30um')
        self.assertEqual(options['inductor_width'], '20um')
        self.assertEqual(options['pad_width'], '455um')
        self.assertEqual(options['pad_height'], '90um')
        self.assertEqual(options['pocket_width'], '650um')
        self.assertEqual(options['pocket_height'], '650um')

        self.assertEqual(len(options['_default_connection_pads']), 12)
        self.assertEqual(options['_default_connection_pads']['pad_gap'], '15um')
        self.assertEqual(options['_default_connection_pads']['pad_width'],
                         '125um')
        self.assertEqual(options['_default_connection_pads']['pad_height'],
                         '30um')
        self.assertEqual(options['_default_connection_pads']['pad_cpw_shift'],
                         '0um')
        self.assertEqual(options['_default_connection_pads']['pad_cpw_extent'],
                         '25um')
        self.assertEqual(options['_default_connection_pads']['cpw_width'],
                         '10um')
        self.assertEqual(options['_default_connection_pads']['cpw_gap'], '6um')
        self.assertEqual(options['_default_connection_pads']['cpw_extend'],
                         '100um')
        self.assertEqual(options['_default_connection_pads']['pocket_extent'],
                         '5um')
        self.assertEqual(options['_default_connection_pads']['pocket_rise'],
                         '0um')
        self.assertEqual(options['_default_connection_pads']['loc_W'], '+1')
        self.assertEqual(options['_default_connection_pads']['loc_H'], '+1')

    def test_qlibrary_transmon_pocket_teeth_options(self):
        """Test that default_options of transmon_pocket_teeth were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        transmon_pocket_teeth = TransmonPocketTeeth(design, 'my_name')
        options = transmon_pocket_teeth.default_options

        self.assertEqual(len(options), 10)
        self.assertEqual(options['pad_gap'], '30um')
        self.assertEqual(options['inductor_width'], '20um')
        self.assertEqual(options['pad_width'], '400um')
        self.assertEqual(options['pad_height'], '90um')
        self.assertEqual(options['pocket_width'], '650um')
        self.assertEqual(options['pocket_height'], '650um')
        self.assertEqual(options['coupled_pad_height'], '150um')
        self.assertEqual(options['coupled_pad_width'], '20um')
        self.assertEqual(options['coupled_pad_gap'], '50um')

        self.assertEqual(len(options['_default_connection_pads']), 12)
        self.assertEqual(options['_default_connection_pads']['pad_gap'], '15um')
        self.assertEqual(options['_default_connection_pads']['pad_width'],
                         '20um')
        self.assertEqual(options['_default_connection_pads']['pad_height'],
                         '150um')
        self.assertEqual(options['_default_connection_pads']['pad_cpw_shift'],
                         '0um')
        self.assertEqual(options['_default_connection_pads']['pad_cpw_extent'],
                         '25um')
        self.assertEqual(options['_default_connection_pads']['cpw_width'],
                         '10um')
        self.assertEqual(options['_default_connection_pads']['cpw_gap'], '6um')
        self.assertEqual(options['_default_connection_pads']['cpw_extend'],
                         '100um')
        self.assertEqual(options['_default_connection_pads']['pocket_extent'],
                         '5um')
        self.assertEqual(options['_default_connection_pads']['pocket_rise'],
                         '0um')
        self.assertEqual(options['_default_connection_pads']['loc_W'], '+1')
        self.assertEqual(options['_default_connection_pads']['loc_H'], '+1')

    def test_qlibrary_squid_loop_options(self):
        """Test that default_options of squid loop were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        squid_loop = SQUID_LOOP(design, 'my_name')
        options = squid_loop.default_options

        self.assertEqual(len(options), 15)
        self.assertEqual(options['plate1_width'], '5.5um')
        self.assertEqual(options['plate1_height'], '40um')
        self.assertEqual(options['plate1_pos_x'], '0')
        self.assertEqual(options['plate1_pos_y'], '0')
        self.assertEqual(options['squid_gap'], '10um')
        self.assertEqual(options['segment_a_length'], '10um')
        self.assertEqual(options['segment_a_width'], '1um')
        self.assertEqual(options['JJ_gap'], '0.5um')
        self.assertEqual(options['segment_b_length'], '5um')
        self.assertEqual(options['segment_b_width'], '1um')
        self.assertEqual(options['segment_c_width'], '1um')
        self.assertEqual(options['segment_d_length'], '10um')
        self.assertEqual(options['segment_d_width'], '2um')
        self.assertEqual(options['plate2_width'], '6um')
        self.assertEqual(options['plate2_height'], '30um')

    def test_qlibrary_tunable_coupler_01_options(self):
        """Test that default_options of tunable_coupler_01 were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        tunable_coupler = TunableCoupler01(design, 'my_name')
        options = tunable_coupler.default_options

        self.assertEqual(len(options), 13)
        self.assertEqual(options['c_width'], '400um')
        self.assertEqual(options['l_width'], '20um')
        self.assertEqual(options['l_gap'], '10um')
        self.assertEqual(options['a_height'], '60um')
        self.assertEqual(options['cp_height'], '15um')
        self.assertEqual(options['cp_arm_length'], '30um')
        self.assertEqual(options['cp_arm_width'], '6um')
        self.assertEqual(options['cp_gap'], '6um')
        self.assertEqual(options['cp_gspace'], '3um')
        self.assertEqual(options['fl_width'], '5um')
        self.assertEqual(options['fl_gap'], '3um')
        self.assertEqual(options['fl_length'], '10um')
        self.assertEqual(options['fl_ground'], '2um')

    def test_qlibrary_transmon_cross_options(self):
        """Test that default options of transmon_cross in transmon_cross.py
        were not accidentally changed."""
        # Setup expected test results
        _design = designs.DesignPlanar()
        _transmon_cross = transmon_cross.TransmonCross(_design, 'my_name')
        _options = _transmon_cross.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(_options), 5)
        self.assertEqual(_options['cross_width'], '20um')
        self.assertEqual(_options['cross_length'], '200um')
        self.assertEqual(_options['cross_gap'], '20um')

        self.assertEqual(len(_options['_default_connection_pads']), 8)
        self.assertEqual(_options['_default_connection_pads']['connector_type'],
                         '0')
        self.assertEqual(_options['_default_connection_pads']['claw_length'],
                         '30um')
        self.assertEqual(_options['_default_connection_pads']['ground_spacing'],
                         '5um')
        self.assertEqual(_options['_default_connection_pads']['claw_width'],
                         '10um')
        self.assertEqual(_options['_default_connection_pads']['claw_gap'],
                         '6um')
        self.assertEqual(
            _options['_default_connection_pads']['claw_cpw_length'], '40um')
        self.assertEqual(_options['_default_connection_pads']['claw_cpw_width'],
                         '10um')
        self.assertEqual(
            _options['_default_connection_pads']['connector_location'], '0')

    def test_qlibrary_transmon_pocket_options(self):
        """Test that default options of transmon_pocket in transmon_pocket.py
        were not accidentally changed."""
        # Setup expected test results
        _design = designs.DesignPlanar()
        _transmon_pocket = transmon_pocket.TransmonPocket(_design, 'my_name')
        _options = _transmon_pocket.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(_options), 7)
        self.assertEqual(_options['pad_gap'], '30um')
        self.assertEqual(_options['inductor_width'], '20um')
        self.assertEqual(_options['pad_width'], '455um')
        self.assertEqual(_options['pad_height'], '90um')
        self.assertEqual(_options['pocket_width'], '650um')
        self.assertEqual(_options['pocket_height'], '650um')

        self.assertEqual(len(_options['_default_connection_pads']), 12)
        self.assertEqual(_options['_default_connection_pads']['pad_gap'],
                         '15um')
        self.assertEqual(_options['_default_connection_pads']['pad_width'],
                         '125um')
        self.assertEqual(_options['_default_connection_pads']['pad_height'],
                         '30um')
        self.assertEqual(_options['_default_connection_pads']['pad_cpw_shift'],
                         '5um')
        self.assertEqual(_options['_default_connection_pads']['pad_cpw_extent'],
                         '25um')
        self.assertEqual(_options['_default_connection_pads']['cpw_width'],
                         'cpw_width')
        self.assertEqual(_options['_default_connection_pads']['cpw_gap'],
                         'cpw_gap')
        self.assertEqual(_options['_default_connection_pads']['cpw_extend'],
                         '100um')
        self.assertEqual(_options['_default_connection_pads']['pocket_extent'],
                         '5um')
        self.assertEqual(_options['_default_connection_pads']['pocket_rise'],
                         '65um')
        self.assertEqual(_options['_default_connection_pads']['loc_W'], '+1')
        self.assertEqual(_options['_default_connection_pads']['loc_H'], '+1')

    def test_qlibrary_transmon_pocket_cl_options(self):
        """Test that default options of transmon_pocket_cl in
        transmon_pocket_cl.py were not accidentally changed."""
        # Setup expected test results
        _design = designs.DesignPlanar()
        _transmon_pocket_cl = transmon_pocket_cl.TransmonPocketCL(
            _design, 'my_name')
        _options = _transmon_pocket_cl.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(_options), 7)
        self.assertEqual(_options['make_CL'], True)
        self.assertEqual(_options['cl_gap'], '6um')
        self.assertEqual(_options['cl_width'], '10um')
        self.assertEqual(_options['cl_length'], '20um')
        self.assertEqual(_options['cl_ground_gap'], '6um')
        self.assertEqual(_options['cl_pocket_edge'], '0')
        self.assertEqual(_options['cl_off_center'], '50um')

    def test_qlibrary_cpw_finger_cap_options(self):
        """Test that default options of CapNInterdigital in cap_n_interdigital.py were not
        accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        finger_cap = CapNInterdigital(design, 'my_name')
        options = finger_cap.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 10)
        self.assertEqual(options['north_width'], '10um')
        self.assertEqual(options['north_gap'], '6um')
        self.assertEqual(options['south_width'], '10um')
        self.assertEqual(options['south_gap'], '6um')
        self.assertEqual(options['cap_width'], '10um')
        self.assertEqual(options['cap_gap'], '6um')
        self.assertEqual(options['cap_gap_ground'], '6um')
        self.assertEqual(options['finger_length'], '20um')
        self.assertEqual(options['finger_count'], '5')
        self.assertEqual(options['cap_distance'], '50um')

    def test_qlibrary_cpw_t_finger_cap_options(self):
        """Test that default options of CapNInterdigitalTee in cap_n_interdigital_tee.py were not
        accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        t_finger_cap = CapNInterdigitalTee(design, 'my_name')
        options = t_finger_cap.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 9)
        self.assertEqual(options['prime_width'], '10um')
        self.assertEqual(options['prime_gap'], '6um')
        self.assertEqual(options['second_width'], '10um')
        self.assertEqual(options['second_gap'], '6um')
        self.assertEqual(options['cap_gap'], '6um')
        self.assertEqual(options['cap_width'], '10um')
        self.assertEqual(options['finger_length'], '20um')
        self.assertEqual(options['finger_count'], '5')
        self.assertEqual(options['cap_distance'], '50um')

    def test_qlibrary_cpw_t_options(self):
        """Test that default options of LineTee in line_tee.py were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        cpw_t = LineTee(design, 'my_name')
        options = cpw_t.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 5)
        self.assertEqual(options['prime_width'], '10um')
        self.assertEqual(options['prime_gap'], '6um')
        self.assertEqual(options['second_width'], '10um')
        self.assertEqual(options['second_gap'], '6um')
        self.assertEqual(options['t_length'], '50um')

    def test_qlibrary_cpw_hanger_t_options(self):
        """Test that default options of CoupledLineTee in coupled_line_tee.py were not
        accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        hanger_t = CoupledLineTee(design, 'my_name')
        options = hanger_t.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 10)
        self.assertEqual(options['prime_width'], '10um')
        self.assertEqual(options['prime_gap'], '6um')
        self.assertEqual(options['second_width'], '10um')
        self.assertEqual(options['second_gap'], '6um')
        self.assertEqual(options['coupling_space'], '3um')
        self.assertEqual(options['coupling_length'], '100um')
        self.assertEqual(options['down_length'], '100um')
        self.assertEqual(options['fillet'], '25um')
        self.assertEqual(options['mirror'], False)
        self.assertEqual(options['open_termination'], True)

    def test_qlibrary_resonator_rectangle_spiral_options(self):
        """Test that default options of ResonatorCoilRect in
        resonator_coil_rect.py were not accidentally changed."""
        # Setup expected test results
        design = designs.DesignPlanar()
        resonator_rectangle_spiral = ResonatorCoilRect(design, 'my_name')
        options = resonator_rectangle_spiral.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 6)
        self.assertEqual(options['n'], '3')
        self.assertEqual(options['length'], '2000um')
        self.assertEqual(options['line_width'], '1um')
        self.assertEqual(options['height'], '40um')
        self.assertEqual(options['gap'], '4um')
        self.assertEqual(options['coupler_distance'], '10um')

    def test_qlibrary_route_anchors_options(self):
        """Test that default options of RouteAnchors in anchored_path.py were
        not accientally changed."""
        # Setup expected test results
        route_anchors = RouteAnchors
        options = route_anchors.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 2)
        self.assertEqual(options['anchors'], {})
        self.assertEqual(len(options['advanced']), 1)
        self.assertEqual(options['advanced']['avoid_collision'], 'false')

    def test_qlibrary_route_pathfinder_options(self):
        """Test that default options of RoutePathfinder in pathfinder.py were
        not accidentally changed."""
        # Setup expected test results
        route_pathfinder = RoutePathfinder
        options = route_pathfinder.default_options

        # Test all elements of the result data against expected data
        self.assertEqual(len(options), 2)
        self.assertEqual(options['step_size'], '0.25mm')
        self.assertEqual(len(options['advanced']), 1)
        self.assertEqual(options['advanced']['avoid_collision'], 'true')

    def test_qlibrary_launch_v1_options(self):
        """Test that default options of LaunchpadWirebond in launchpad_wb.py
        were not accidentally changed."""
        design = designs.DesignPlanar()
        launch_v1 = LaunchpadWirebond(design, 'my_name')
        options = launch_v1.default_options

        self.assertEqual(len(options), 7)
        self.assertEqual(options['trace_width'], 'cpw_width')
        self.assertEqual(options['trace_gap'], 'cpw_gap')
        self.assertEqual(options['lead_length'], '25um')
        self.assertEqual(options['pad_width'], '80um')
        self.assertEqual(options['pad_height'], '80um')
        self.assertEqual(options['pad_gap'], '58um')
        self.assertEqual(options['taper_height'], '122um')

    def test_qlibrary_launch_v2_options(self):
        """Test that default options of LaunchpadWirebondCoupled in
        launchpad_wb_coupled.py were not accidentally changed."""
        design = designs.DesignPlanar()
        launch_v2 = LaunchpadWirebondCoupled(design, 'my_name')
        options = launch_v2.default_options

        self.assertEqual(len(options), 4)
        self.assertEqual(options['trace_width'], 'cpw_width')
        self.assertEqual(options['trace_gap'], 'cpw_gap')
        self.assertEqual(options['coupler_length'], '62.5um')
        self.assertEqual(options['lead_length'], '25um')

    def test_qlibrary_cap_three_fingers(self):
        """Test that default options of Cap3Interdigital were not accidentally
        changed."""
        design = designs.DesignPlanar()
        cap_three_fingers = Cap3Interdigital(design, 'my_name')
        options = cap_three_fingers.default_options

        self.assertEqual(len(options), 4)
        self.assertEqual(options['trace_width'], '10um')
        self.assertEqual(options['finger_length'], '65um')
        self.assertEqual(options['pocket_buffer_width_x'], '10um')
        self.assertEqual(options['pocket_buffer_width_y'], '30um')

    def test_qlibrary_qroute_options(self):
        """Test that default options of QRoute were not accidentally changed."""
        design = designs.DesignPlanar()
        my_qroute = qroute.QRoute(design, name='test_qroute', options={})
        options = my_qroute.default_options

        self.assertEqual(len(options), 5)
        self.assertEqual(options['fillet'], '0')
        self.assertEqual(options['total_length'], '7mm')
        self.assertEqual(options['trace_width'], 'cpw_width')

        self.assertEqual(len(options['pin_inputs']), 2)
        self.assertEqual(len(options['pin_inputs']['start_pin']), 2)
        self.assertEqual(len(options['pin_inputs']['end_pin']), 2)
        self.assertEqual(options['pin_inputs']['start_pin']['component'], '')
        self.assertEqual(options['pin_inputs']['start_pin']['pin'], '')
        self.assertEqual(options['pin_inputs']['end_pin']['component'], '')
        self.assertEqual(options['pin_inputs']['end_pin']['pin'], '')

        self.assertEqual(len(options['lead']), 4)
        self.assertEqual(options['lead']['start_straight'], '0mm')
        self.assertEqual(options['lead']['end_straight'], '0mm')
        self.assertEqual(options['lead']['start_jogged_extension'], '')
        self.assertEqual(options['lead']['end_jogged_extension'], '')

    def test_qlibrary_jj_dolan_options(self):
        """Test the default options of JJ_Dolan were not accidentially changed."""
        design = designs.DesignPlanar()
        my_jj_dolan = jj_dolan(design, name='test_jj_dolan', options={})
        options = my_jj_dolan.default_options

        self.assertEqual(len(options), 10)
        self.assertEqual(options['JJ_pad_lower_width'], '25um')
        self.assertEqual(options['JJ_pad_lower_height'], '10um')
        self.assertEqual(options['JJ_pad_lower_pos_x'], '0')
        self.assertEqual(options['JJ_pad_lower_pos_y'], '0')
        self.assertEqual(options['finger_lower_width'], '1um')
        self.assertEqual(options['finger_lower_height'], '20um')
        self.assertEqual(options['extension'], '1um')
        self.assertEqual(options['offset'], '2um')
        self.assertEqual(options['second_metal_length'], '5um')
        self.assertEqual(options['second_metal_width'], '1um')

    def test_qlibrary_jj_manhattan_options(self):
        """Test the default options of JJ_Manhattan were not accidentially changed."""
        design = designs.DesignPlanar()
        my_jj_manhattan = jj_manhattan(design,
                                       name='test_jj_manhattan',
                                       options={})
        options = my_jj_manhattan.default_options

        self.assertEqual(len(options), 7)
        self.assertEqual(options['JJ_pad_lower_width'], '25um')
        self.assertEqual(options['JJ_pad_lower_height'], '10um')
        self.assertEqual(options['JJ_pad_lower_pos_x'], '0')
        self.assertEqual(options['JJ_pad_lower_pos_y'], '0')
        self.assertEqual(options['finger_lower_width'], '1um')
        self.assertEqual(options['finger_lower_height'], '20um')
        self.assertEqual(options['extension'], '1um')


if __name__ == '__main__':
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Dict tree base."""
# pylint: disable=invalid-name

import ast
from typing import Union, TYPE_CHECKING

from PySide6 import QtCore
from PySide6.QtCore import QAbstractItemModel, QModelIndex, QTimer, Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import (QWidget, QTreeView)
# from .... import logger

if TYPE_CHECKING:
    from ....designs.design_base import QDesign
    from ....qlibrary.core import QComponent
    from ...main_window import MetalGUI

__all__ = ['get_nested_dict_item', 'parse_param_from_str']

KEY, NODE = range(2)

# List of children given as [(childname_0, childnode_0), (childname_1, childnode_1), ...]
# Where childname_i corresponds to KEY and childnode_i corresponds to NODE


def get_nested_dict_item(dic: dict, key_list: list):
    """Get a nested dictionary item. If key_list is empty, return dic itself.

    Args:
        dic (dict): Dictionary of items
        key_list (list): List of keys

    Returns:
        dict: Nested dictionary

    .. code-block:: python
        :linenos:

        myDict = Dict(aa=Dict(x1={'dda':34},y1='Y',z='10um'),
            bb=Dict(x2=5,y2='YYYsdg',z='100um'))
        key_list = ['aa', 'x1', 'dda']
        [get_dict_item](myDict, key_list)
        returns 34
    """
    if key_list:
        for k in key_list:
            dic = dic[k]
    return dic


class BranchNode:
    """A BranchNode object has a nonzero number of child nodes. These child
    nodes can be either BranchNodes or LeafNodes.

    It is uniquely defined by its name, parent node, and list of
    children. The list of children consists of tuple pairs of the form
    (nodename_i, node_i), where the former is the name of the child node
    and the latter is the child node itself. KEY (=0) and NODE (=1)
    identify their respective positions within each tuple pair.
    """

    def __init__(self, name: str, parent=None, data: dict = None):
        """
        Args:
            name (str): Name of this branch
            parent ([type]): The parent.  Defaults to None.
            data (dict): Node data.  Defaults to None.
        """
        super(BranchNode, self).__init__()
        self.name = name
        self.parent = parent
        self.children = []
        self._data = data  # dictionary containing the actual data

    def __len__(self):
        """Gets the number of children.

        Returns:
            int: The number of children this node has
        """
        return len(self.children)

    def provideName(self):
        """Gets the name.

        Returns:
            str: The nodes name
        """
        return self.name  # identifier for BranchNode

    def childAtRow(self, row: int):
        """Gets the child at the given row.

        Args:
            row (int): The row

        Returns:
            Node: The node at the row
        """
        if 0 <= row < len(self.children):
            return self.children[row][NODE]

    def rowOfChild(self, child):
        """Gets the row of the given child.

        Args:
            child (Node): The child

        Returns:
            int: Row of the given child.  -1 is returned if the child is not found.
        """
        for i, (_, childnode) in enumerate(self.children):
            if child == childnode:
                return i
        return -1

    def childWithKey(self, key):
        """Gets the child with the given key.

        Args:
            key (str): The key

        Returns:
            Node: The child with the same name as the given key.
            None is returned if the child is not found
        """
        for childname, childnode in self.children:
            if key == childname:
                return childnode
        return None

    def insertChild(self, child):
        """Insert the given child.

        Args:
            child (Node): The child
        """
        child.parent = self
        self.children.append((child.provideName(), child))

    def hasLeaves(self):
        """Do I have leaves?

        Returns:
            bool: True if there are leaves, False otherwise
        """
        if not self.children:
            return False
        return isinstance(self.children[KEY][NODE], LeafNode)


class LeafNode:
    """A LeafNode object has no children but consists of a key-value pair,
    denoted by label and value, respectively.

    It is uniquely identified by its root-to-leaf path, which is a list
    of keys whose positions denote their nesting depth (shallow to
    deep).
    """

    def __init__(self, label: str, parent=None, path=None):
        """
        Args:
            label (str): Label for the leaf node
            parent (Node): The parent.  Defaults to None.
            path (list): Node path.  Defaults to None.
        """
        super(LeafNode, self).__init__()
        self.path = path or []
        self.parent = parent
        self.label = label

    @property
    def value(self):
        """Returns the value."""
        return get_nested_dict_item(self.parent._data, self.path)

    # @value.setter
    # def value(self, newvalue):
    #     self._value = newvalue

    def provideName(self):
        """Get the label.

        Returns:
            str: The lable of the leaf node - this is *not* the value.
        """
        return self.label  # identifier for LeafNode (note: NOT value!)


class QTreeModel_Base(QAbstractItemModel):
    """Tree model for a general hierarchical dataset.

    This class extends the `QAbstractItemModel` class.
    It is part of the model-view-controller (MVC) architecture; see
    https://doc.qt.io/qt-5/qabstractitemmodel.html

    Access using ``gui.component_window.model``.
    """

    __refreshtime = 500  # 0.5 second refresh time

    # NOTE: __init__ takes in design as extra parameter compared to table_model_options!

    def __init__(self, parent: QWidget, gui: 'MetalGUI', view: QTreeView,
                 child: str):
        """Editable table with drop-down rows for a generic options menu.
        Organized as a tree model where child nodes are more specific
        properties of a given parent node.

        Args:
            parent (QWidget): The parent widget
            gui (MetalGUI): The main user interface
            view (QTreeView): View corresponding to a tree structure
            child (str): Name of child class, eg 'component' or 'GDS renderer'
        """
        super().__init__(parent=parent)
        self.logger = gui.logger
        self._gui = gui
        self._row_count = -1
        self._view = view
        self.optionstype = child
        if self.optionstype == 'component':
            self._component_widget = parent

        self.root = BranchNode('')
        self.headers = ['Name', 'Value']
        self.paths = []

        self._start_timer()
        self.load()  # handles beginResetModel and endResetModel

    @property
    def gui(self):
        """Returns the GUI."""
        return self._gui

    @property
    def design(self) -> 'QDesign':
        """Returns the QDesign."""
        return self._gui.design

    @property
    def component(self) -> 'QComponent':
        """Returns the component if this is the components options menu."""
        if self.optionstype == 'component':
            return self._component_widget.component
        return None

    def _start_timer(self):
        """Start and continuously refresh timer in background to keep the total
        number of rows up to date."""
        self.timer = QTimer(self)
        self.timer.start(self.__refreshtime)
        self.timer.timeout.connect(self.auto_refresh)

    def auto_refresh(self):
        """Check to see if the total number of rows has been changed.

        If so, completely rebuild the model and tree.
        """
        # TODO: Check if new nodes have been added; if so, rebuild model.
        new_row_count = self.rowCount(self.createIndex(0, 0))
        if self._row_count != new_row_count:
            # Wrap the reset logic in beginResetModel and endResetModel
            self.beginResetModel()
            try:

                # When a model is reset it should be considered that all
                # information previously retrieved from it is invalid.
                # This includes but is not limited to the rowCount() and
                # columnCount(), flags(), data retrieved through data(), and roleNames().
                # This will loose the current selection.
                # self.modelReset.emit()

                self._row_count = new_row_count
            finally:
                self.endResetModel()
            if self._view:
                self._view.autoresize_columns()

    def refresh(self):
        """Force refresh.

        Completely rebuild the model and tree.
        """
        # self.beginResetModel()  # load handles refresh, cant nest
        # try:
        self.load(
        )  # rebuild the tree; handles beginResetModel and endResetModel
        parent_index = self.createIndex(0, 0, self.root)
        self._row_count = self.rowCount(parent_index)
        # finally:
        #     # self.endResetModel()

    def getPaths(self, curdict: dict, curpath: list):
        """Recursively finds and saves all root-to-leaf paths in model."""
        for k, v in curdict.items():
            if isinstance(v, dict):
                self.getPaths(v, curpath + [k])
            else:
                self.paths.append(curpath + [k, v])

    def load(self):
        """Builds a tree from a dictionary (self.data_dict)"""
        if (self.optionstype == 'component') and (not self.component):
            return

        self.beginResetModel()

        # Set the data dict reference of the root node. The root node doesn't have a name.
        self.root._data = self.data_dict

        # Clear existing tree paths if any
        self.paths.clear()
        self.root.children.clear()

        # Construct the paths -> sets self.paths
        self.getPaths(self.data_dict, [])

        for path in self.paths:
            root = self.root
            branch = None
            # Combine final label and value for leaf node,
            # so stop at 2nd to last element of each path
            for key in path[:-2]:
                # Look for childnode with the name 'key'. If it's not found, create a new branch.
                branch = root.childWithKey(key)
                if not branch:
                    branch = BranchNode(key, data=self.data_dict)
                    root.insertChild(branch)
                root = branch
            # If a LeafNode resides in the outermost dictionary, the above for loop is bypassed.
            # [Note: This happens when the root-to-leaf path length is just 2.]
            # In this case, add the LeafNode right below the master root.
            # Otherwise, add the LeafNode below the final branch.
            if not branch:
                root.insertChild(LeafNode(path[-2], root, path=path[:-1]))
            else:
                branch.insertChild(LeafNode(path[-2], branch, path=path[:-1]))

        # Emit a signal since the model's internal state
        # (e.g. persistent model indexes) has been invalidated.
        self.endResetModel()

    def rowCount(self, parent: QModelIndex):
        """Get the number of rows.

        Args:
            parent (QModelIndex): The parent

        Returns:
            int: The number of rows
        """
        # Count number of child nodes belonging to the parent.
        node = self.nodeFromIndex(parent)
        if (node is None) or isinstance(node, LeafNode):
            return 0
        return len(node)

    def columnCount(self, parent: QModelIndex):  # pylint: disable=unused-argument
        """Get the number of columns.

        Args:
            parent (QModelIndex): The parent

        Returns:
            int: The number of columns
        """
        return len(self.headers)

    def data(self, index: QModelIndex, role: Qt.ItemDataRole = Qt.DisplayRole):
        """Gets the node data.

        Args:
            index (QModelIndex): Index to get data for
            role (Qt.ItemDataRole): The role.  Defaults to Qt.DisplayRole.

        Returns:
            object: Fetched data
        """
        if not index.isValid():
            return None

        # The data in a form suitable for editing in an editor. (QString)
        if role == Qt.EditRole:
            return self.data(index, Qt.DisplayRole)

        # Bold the first
        if (role == Qt.FontRole) and (index.column() == 0):
            font = QFont()
            font.setBold(True)
            return font

        if role == Qt.TextAlignmentRole:
            return int(Qt.AlignTop | Qt.AlignLeft)

        if role == Qt.DisplayRole:
            node = self.nodeFromIndex(index)
            if node:
                # the first column is either a leaf key or a branch
                # the second column is always a leaf value or for a branch is ''.
                if isinstance(node, BranchNode):
                    # Handle a branch (which is a nested subdictionary, which can be expanded)
                    if index.column() == 0:
                        return node.name
                    return ''
                # We have a leaf
                elif index.column() == 0:
                    return str(node.label)  # key
                elif index.column() == 1:
                    return str(node.value)  # value
                else:
                    return None

        return None

    def setData(self,
                index: QModelIndex,
                value,
                role: Qt.ItemDataRole = Qt.EditRole) -> bool:
        """Set the LeafNode value and corresponding data entry to value.
        Returns true if successful; otherwise returns false. The dataChanged()
        signal should be emitted if the data was successfully set.

        Args:
            index (QModelIndex): The index
            value: The value
            role (Qt.ItemDataRole): The role of the data..  Defaults to Qt.EditRole.

        Returns:
            bool: True if successful, False otherwise
        """

        if not index.isValid():
            return False

        elif role == QtCore.Qt.EditRole:

            if index.column() == 1:
                node = self.nodeFromIndex(index)

                if isinstance(node, LeafNode):
                    value = str(value)  # new value
                    old_value = node.value  # option value

                    if old_value == value:
                        return False

                    # Set the value of an option when the new value is different
                    else:
                        dic = self.data_dict  # option dict
                        lbl = node.label  # option key

                        self.logger.info(
                            f'Setting {self.optionstype} option {lbl:>10s}:'
                            f' old value={old_value}; new value={value};')

                        ##### Parse value if not str ##############################
                        # Somewhat legacy code for extended handling of non string options
                        # These days we tend to have all options be strings,
                        # so not so releavnt, but keep here for now
                        # to allow extended use in te future
                        if not isinstance(old_value, str):
                            processed_value, used_ast = parse_param_from_str(
                                value)
                            self.logger.info(f'  Used paring:  Old value type={type(old_value)}; '\
                                             f'New value type={type(processed_value)};'\
                                             f'  New value={processed_value};'\
                                             f'; Used ast={used_ast}')
                            value = processed_value
                        #################################################

                        if node.path:  # if nested option
                            for x in node.path[:-1]:
                                dic = dic[x]
                            dic[node.path[-1]] = value
                        else:  # if top-level option
                            dic[lbl] = value
                        if self.optionstype == 'component':
                            self.component.rebuild()
                            self.gui.refresh()
                        return True
        return False

    def headerData(self, section: int, orientation: Qt.Orientation,
                   role: Qt.ItemDataRole):
        """Set the headers to be displayed.

        Args:
            section (int): Section number
            orientation (Qt.Orientation): The orientation
            role (Qt.ItemDataRole): The role

        Returns:
            object: Header data
        """
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                if 0 <= section < len(self.headers):
                    return self.headers[section]

            elif role == Qt.FontRole:
                font = QFont()
                font.setBold(True)
                return font

        return None

    def index(self, row: int, column: int, parent: QModelIndex):
        """Return my index.

        Args:
            row (int): The row
            column (int): The column
            parent (QModelIndex): The parent

        Returns:
            int: internal index
        """
        assert self.root
        branch = self.nodeFromIndex(parent)
        assert branch is not None
        # The third argument is the internal index.
        return self.createIndex(row, column, branch.childAtRow(row))

    def parent(self, child):
        """Gets the parent index of the given node.

        Args:
            child (node): The child

        Returns:
            int: The index
        """
        node = self.nodeFromIndex(child)
        if node is None:
            return QModelIndex()
        parent = node.parent
        if parent is None:
            return QModelIndex()
        grandparent = parent.parent
        if grandparent is None:
            return QModelIndex()
        row = grandparent.rowOfChild(parent)
        assert row != -1
        return self.createIndex(row, 0, parent)

    def nodeFromIndex(self, index: QModelIndex) -> Union[BranchNode, LeafNode]:
        """Utility method we define to get the node from the index.

        Args:
            index (QModelIndex): The model index

        Returns:
            Union[BranchNode, LeafNode]: Returns the node, which can either be
            a branch (for a dict) or a leaf (for an option key value pairs)
        """
        if index.isValid():
            # The internal pointer will return the leaf or branch node under the given parent.
            return index.internalPointer()
        return self.root

    def flags(self, index: QModelIndex):
        """Determine how user may interact with each cell in the table.

        Returns:
            list: List of flags
        """
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled
        if index.column() == 1:
            node = self.nodeFromIndex(index)
            if isinstance(node, LeafNode):
                return flags | Qt.ItemIsEditable

        return flags


def parse_param_from_str(text):
    """Attempt to parse a value from a string using ast.

    Args:
        text (str): String to parse

    Return:
        tuple: value, used_ast

    Raises:
        Exception: An error occurred
    """
    text = str(text).strip()
    value = text
    used_ast = False
    try:  # crude way to handle list and values
        value = ast.literal_eval(text)
        used_ast = True
    except Exception:  #  as exception
        pass
        # print(exception)
    return value, used_ast



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

from abc import abstractmethod, ABC
import inspect
from typing import Any, Union

from copy import deepcopy
from qiskit_metal import logger, Dict
from qiskit_metal.analyses.sweep_and_optimize.sweeper import Sweeper


class QAnalysis(ABC):
    """`QAnalysis` is the core class from which all Metal analysis classes are derived.

    The class defines the user interface for working with analysis.

    For front-end user:
        * Manipulates the setup dictionary to fine-tune the analysis parameters.

    For creator user:
        * Creates a class that inherits this or the QSimulation classes.
        * Implements the `run` and `run_***` methods to define the analysis procedures.
        * Defines the methods to communicate with the QRenderers. (see QSimulation)
        * Defines default_setup, which describes the parameteric interface to the analysis class.

    Default Setup:
        Nested default setup parameters can be overwritten with the setup_update method,
        or directly by using the dot operator enabled by the Dict type.
    """

    default_setup = Dict()
    """Default setup."""

    data_labels = list()
    """Default data labels."""

    def __init__(self, *args, **kwargs):
        """Creates a new Analysis object with a setup derived from the default_setup Dict.
        """
        super().__init__(*args, **kwargs)
        self._setup, self._supported_data_labels = self._gather_from_all_children(
        )

        # first self.clear_data()
        self._variables = Dict()

        # Keep a reference to Sweeper class.
        self._sweeper = None

    def _initialize_sweep(self):
        """Create a new instance of Sweeper
        """
        self._sweeper = Sweeper(self)

    @property
    def logger(self):
        """Returns the logger."""
        return logger

    @abstractmethod
    def run(self, *args, **kwargs):
        """Abstract method. Must be implemented by the subclass.
        Use as an alias for the main analysis method. Call the main analysis method run_***().
        Subclass implementation of run() must:
        * always call the main analysis method(s) named `run_***()`.
        * never implement the analysis, which should be instead in the method run_***().

        `run_***()` must always call self.save_run_args(`*args`, `**kwargs`) to save the inputs.

        Example:

            ::

                def run(self, *args, **kwargs):
                    self.run_sim(*args, **kwargs)
                def run_sim(self, *args, **kwargs):
                    self.save_run_args(*args, **kwargs)
                    <......>

        Make sure the name of the method `run_***()` does not conflict with that of another
        QAnalysis subclass, in case you expect them to be both inherited from the same subclass.
        """

    def run_sweep(self, *args, **kwargs):
        """User requests sweeper based on arguments from
        Sweeper.run_sweep().
        """
        if not self._sweeper:
            self._initialize_sweep()

        all_sweep, return_code = self._sweeper.run_sweep(*args, **kwargs)
        return all_sweep, return_code

    def save_run_args(self, **kwargs):
        """Intended to be used to store the kwargs passed to the run() method,
        for repeatability and for later identification of the QAnalysis instance.
        """
        self._setup.run = None
        self._setup.run = Dict(kwargs)

    def set_data(self, data_name: str, data: Any):
        """Stores data in a structure for later retrieval.
        Could be output, intermediate or even input data.
        Current implementation uses Dict()

        Args:
            data_name (str): Label for the data. Used a storage key.
            data (Any): Free format
        """
        if data_name not in self._supported_data_labels:
            self.logger.warning(
                'No %s in the list of supported variables. The variable will still be added. '
                'However, make sure this was not a typo', {data_name})
        self._variables[data_name] = data

    def get_data(self, data_name: str = None):
        """Retrieves the analysis module data.
        Returns `None` if nothing is found.

        Args:
            data_name (str, optional): Label to query for data. If not specified, the
                entire dictionary is returned. Defaults to None.

        Returns:
            Any: The data associated with the label, or the entire list of labels and data.
        """
        if data_name is None:
            return self._variables
        return self._variables[data_name]

    def get_data_labels(self) -> list:
        """Retrieves the list of data labels currently set.
        Returns `None` if nothing is found.

        Returns:
            list: list of data names
        """
        return self._variables.keys()

    @property
    def supported_data(self):
        """Getter: Set that contains the names of the variables supported from the analysis.

        Returns:
            set: list of supported variable names.
        """
        return self._supported_data_labels

    def clear_data(self, data_name: Union[str, list] = None):
        """Clear data. Can optionally specify one or more labels to delete those labels and data.

        Args:
            data_name (Union[str, list], optional): Can list specific labels to clean.
                Defaults to None.
        """
        if data_name is None:
            self._variables = Dict()
        else:
            if isinstance(data_name, str):
                self._variables[data_name] = None
            else:
                for name in data_name:
                    if name in self._variables[name]:
                        del self._variables[name]

    def print_run_args(self):
        """Prints the args and kwargs that were used in the last run() of this Analysis instance.
        """
        print("This analysis object run with the following kwargs:\n"
              f"{self._setup.run}\n")

    @property
    def setup(self):
        """Getter: Dictionary intended to be used to modify the analysis behavior.

        Returns:
            Dict: Current setup.
        """
        return self._setup

    @setup.setter
    def setup(self, setup_dict):
        """Setter: Dictionary intended to be used to modify the analysis behavior. You can only
        pass to this method those settings that are already defined in the default_setup list.
        Non specified keys, will be assigned the default_setup value.

        Args:
            setup_dict (Dict): Define the settings to change. The rest will be set to defaults.
        """
        if isinstance(setup_dict, dict):
            # first reset the setup
            self._setup, _ = self._gather_from_all_children()
            # then apply specified variables
            self.setup_update(**setup_dict)
        else:
            print("The analysis setup has to be defined as a dictionary")

    def setup_update(self, section: str = None, **kwargs):
        """Intended to modify multiple setup settings at once, while retaining previous settings.
        If you intend to change a single setting, the better way is: `setup.setting1 = value`.

        Args:
            section (str): Setup section that contains the setup keys to update.
        """
        if section in self._setup or section is None:
            unsupported_keys = list()
            for k, v in kwargs.items():
                if section is None:
                    s = self._setup
                else:
                    s = self._setup[section]
                if k in s:
                    s[k] = v
                else:
                    unsupported_keys.append(k)
            if unsupported_keys:
                print(
                    f'the parameters {unsupported_keys} are unsupported, so they have been ignored'
                )
        else:
            print(
                f'the section {section} does not exist in this analysis setup')

    @classmethod
    def _gather_from_all_children(cls):
        """From the QAnalysis core class, traverse the child classes to
        gather the default_setup and data_labels for each child class.

        For default_setup: If the same key is found twice, the youngest child will be picked.

        Returns:
            (dict, set): dict = setup, set = supported_data_labels
        """
        setup_from_children = Dict()
        labels_from_children = set()
        parents = inspect.getmro(cls)

        # len-2: base.py is not expected to have default_setup and data_labels.
        for child in parents[len(parents) - 2::-1]:
            # The template default options are in the attribute `default_setup`.
            if hasattr(child, 'default_setup'):
                setup_from_children.update(deepcopy(child.default_setup))
            # The data_labels are in the attribute `data_labels`.
            if hasattr(child, 'data_labels'):
                labels_from_children.update(child.data_labels)

        return setup_from_children, labels_from_children



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './widgets/variable_table/add_delete_table_ui.ui',
# licensing of './widgets/variable_table/add_delete_table_ui.ui' applies.
#
# Created: Sat Jun 19 22:02:31 2021
#      by: pyside6-uic  running on PySide2 5.13.2
#
# WARNING! All changes made in this file will be lost!

from PySide6 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):

    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(344, 206)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred,
                                           QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout_3.setSpacing(0)
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.tableView = RightClickView(self.centralwidget)
        self.tableView.setAutoScroll(False)
        self.tableView.setVerticalScrollMode(
            QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.tableView.setHorizontalScrollMode(
            QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.tableView.setObjectName("tableView")
        self.tableView.horizontalHeader().setVisible(True)
        self.tableView.horizontalHeader().setStretchLastSection(True)
        self.tableView.verticalHeader().setVisible(False)
        self.gridLayout_3.addWidget(self.tableView, 0, 0, 1, 1)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.gridLayout.setContentsMargins(0, -1, -1, -1)
        self.gridLayout.setObjectName("gridLayout")
        self.deleteButton = QtWidgets.QToolButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored,
                                           QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.deleteButton.sizePolicy().hasHeightForWidth())
        self.deleteButton.setSizePolicy(sizePolicy)
        self.deleteButton.setObjectName("deleteButton")
        self.gridLayout.addWidget(self.deleteButton, 0, 1, 1, 1)
        self.addButton = QtWidgets.QToolButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored,
                                           QtWidgets.QSizePolicy.Ignored)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.addButton.sizePolicy().hasHeightForWidth())
        self.addButton.setSizePolicy(sizePolicy)
        self.addButton.setObjectName("addButton")
        self.gridLayout.addWidget(self.addButton, 0, 0, 1, 1)
        self.gridLayout_3.addLayout(self.gridLayout, 1, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar()
        self.menubar.setGeometry(QtCore.QRect(0, 0, 344, 22))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        QtCore.QObject.connect(self.addButton, QtCore.SIGNAL("clicked()"),
                               MainWindow.addRow)
        QtCore.QObject.connect(self.deleteButton, QtCore.SIGNAL("clicked()"),
                               MainWindow.deleteRow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(
            QtWidgets.QApplication.translate("MainWindow", "MainWindow", None,
                                             -1))
        self.deleteButton.setText(
            QtWidgets.QApplication.translate("MainWindow", "Delete variable",
                                             None, -1))
        self.addButton.setText(
            QtWidgets.QApplication.translate("MainWindow", "Add variable", None,
                                             -1))


from .right_click_table_view import RightClickView
